{****************************************************************************

                      Copyright (c) 1996 by Florian Klaempfl

 ****************************************************************************}

{
  This unit implements an asmoutput class for AT&T syntax with
  Intel i386+.

  History:
      30th september 1996:
         + unit started
      15th october 1996:
         + ti386attasmoutput class started
      28th november 1996:
         ! debugging for simple programs

  What's to do:
    o complete op2str array
    o write lines numbers and file names to output file
    o generate debugging informations
    o omit problem of some GNU ASes with pushw $1243
}
unit attasmi3;

  interface

    uses
       globals,systems,errors,cobjects,aasm,i386,strings;

    type
       pi386attasmoutput = ^ti386attasmoutput;

       ti386attasmoutput = object(taasmoutput)
          procedure write_to_file(outfile : pbufferedfile);virtual;
       end;

    { inits the datasegment etc with the output specific list types }
    procedure asmoutlists_init;

    { writes the asmlists to an output file }
    procedure writeasmlists(outfile : pbufferedfile);

  implementation

    const
       _reg2str : array[R_NO..R_SS] of string[4] =
          ('','%eax','%ecx','%edx','%ebx','%esp','%ebp','%esi','%edi',
           '%ax','%cx','%dx','%bx','%sp','%bp','%si','%di',
           '%al','%cl','%dl','%bl',
           '','%cs','%ds','%es','%fs','%gs','%ss');

       op2str : array[A_MOV..A_LAHF] of string[7] =
         ('mov','movz','movs','','add',
          'call','idiv','imul','jmp','lea','mul','neg','not',
          'pop','popal','push','pushal','ret','sub','xchg','xor',
          'fild','cmp','jz','inc','dec','sete','setne','setl',
          'setg','setle','setge','je','jne','jl','jg','jle','jge',
          'or','fld','fadd','fmul','fsub','fdivr','fchs','fld1',
          'fidiv','cltd','jnz','fstp','and','jno','','',
          'enter','leave','cld','movs','rep','shl','shr','bound',
          'jns','js','jo','sar','test',
          'fcom','fcomp','fcompp','fxch','faddp','fmulp','fsubrp','fdivrp',
          'fnsts','sahf','fdivp','fsubp','setc','setnc','jc','jnc',
          'ja','jae','jb','jbe','seta','setae','setb','setbe',
          'aaa','aad','aam','aas','cbw','cdq','clc','cli',
          'clts','cmc','cwd','cwde','daa','das','hlt','iret','lahf');

       opsize2str : array[S_NO..S_WL] of string[2] = ('','b','w','l','bw',
          'bl','wl');

    function getreferencestring(const ref : treference) : string;

      var
         s : string;

      begin
         if ref.isintvalue then
           s:='$'+tostr(ref.offset)
         else
           with ref do
             begin
                { have we a segment prefix ? }
                if segment<>R_DEFAULT_SEG then
                  s:=_reg2str[segment]+':'
                else s:='';

                if assigned(symbol) then
                  s:=s+symbol^;

                if offset<0 then s:=s+tostr(offset)
                  else if (offset>0) then
                    begin
                       if (symbol=nil) then s:=tostr(offset)
                       else s:=s+'+'+tostr(offset);
                    end;
                if (index<>R_NO) and (base=R_NO) then
                  s:=s+'(,'+_reg2str[index]+','+tostr(scalefactor)+')'
                else if (index=R_NO) and (base<>R_NO) then
                  s:=s+'('+_reg2str[base]+')'
                else if (index<>R_NO) and (base<>R_NO) then
                  s:=s+'('+_reg2str[base]+','
                    +_reg2str[index]+','+tostr(scalefactor)+')';
             end;
         getreferencestring:=s;
      end;

    function getopstr(t : byte;o : pointer) : string;

      var
         hs : string;

      begin
         case t of
            top_reg : getopstr:=_reg2str[tregister(o)];
            top_ref : getopstr:=getreferencestring(preference(o)^);
            top_const : getopstr:='$'+tostr(longint(o));
            top_symbol : begin
                            hs[0]:=char(strlen(pchar(pcsymbol(o)^.symbol)));
                            move(pchar(pcsymbol(o)^.symbol)^,hs[1],byte(hs[0]));
                            hs:='$'+hs;
                            if pcsymbol(o)^.offset>0 then
                              hs:=hs+'+'+tostr(pcsymbol(o)^.offset)
                            else if pcsymbol(o)^.offset<0 then
                              hs:=hs+tostr(pcsymbol(o)^.offset);
                            getopstr:=hs;
                         end;
            else internalerror(10001);
         end;
      end;

    function getopstr_jmp(t : byte;o : pointer) : string;

      var
         hs : string;

      begin
         case t of
            top_reg : getopstr_jmp:=_reg2str[tregister(o)];
            top_ref : getopstr_jmp:='*'+getreferencestring(preference(o)^);
            top_const : getopstr_jmp:=tostr(longint(o));
            top_symbol : begin
                            hs[0]:=char(strlen(pchar(pcsymbol(o)^.symbol)));
                            move(pchar(pcsymbol(o)^.symbol)^,hs[1],byte(hs[0]));
                            if pcsymbol(o)^.offset>0 then
                              hs:=hs+'+'+tostr(pcsymbol(o)^.offset)
                            else if pcsymbol(o)^.offset<0 then
                              hs:=hs+tostr(pcsymbol(o)^.offset);
                            getopstr_jmp:=hs;
                         end;
            else internalerror(10001);
         end;
      end;

    function ibm2ascii(const s : string) : string;

      var
         i : integer;
         hs : string;
         b : byte;

      begin
         hs:='';
         for i:=1 to length(s) do
           if ((ord(s[i])>127) or (ord(s[i])<32)) or (s[i]='"') then
             begin
                b:=ord(s[i]);
                hs:=hs+'\'+tostr(b div 64);
                b:=b mod 64;
                hs:=hs+tostr(b div 8);
                b:=b mod 8;
                hs:=hs+tostr(b);
                if (i<length(s)) and
                  (ord(s[i+1])>=48) and  (ord(s[i+1])<=57) then
                  hs:=hs+'","';
             end
           else if s[i]='\' then
             hs:=hs+'\\'
           else hs:=hs+s[i];
         ibm2ascii:=hs;
      end;

{****************************************************************************
                             TI386ATTASMOUTPUT
 ****************************************************************************}

    procedure ti386attasmoutput.write_to_file(outfile : pbufferedfile);

      var
         hp : pai;
         s : string;

      begin
         hp:=pai(first);
         while assigned(hp) do
           begin
              case hp^.typ of
                 { external is ignored }
                 ait_external : ;

                 ait_datablock : begin
                                    if pai_datablock(hp)^.is_global then
                                      outfile^.write_string(#9+'.comm'#9)
                                    else
                                      outfile^.write_string(#9+'.lcomm'#9);
                                    outfile^.write_pchar(pai_datablock(hp)^.name);
                                    outfile^.write_string(','+tostr(pai_datablock(hp)^.size));
                                 end;
                 ait_const_32bit : outfile^.write_string(#9+'.long'#9+tostr(pai_const(hp)^.value));
                 ait_const_16bit : outfile^.write_string(#9+'.short'#9+tostr(pai_const(hp)^.value));
                 ait_const_8bit : outfile^.write_string(#9+'.byte'#9+tostr(pai_const(hp)^.value));
                 ait_const_symbol : begin
                                       outfile^.write_string(#9+'.long'#9);
                                       outfile^.write_pchar(pchar(pai_const(hp)^.value));
                                    end;
                 ait_real_64bit : outfile^.write_string(#9+'.double'#9+double2str(pai_double(hp)^.value));
                 ait_string : begin
                                 outfile^.write_string(#9+'.ascii'#9'"'+ibm2ascii(pai_string(hp)^.str^)+'"');
                              end;
                 ait_label : outfile^.write_string('L'+tostr(pai_label(hp)^.l)+':');
                 ait_labeled_instruction : outfile^.write_string(#9+op2str[pai_labeled386(hp)^.operator]+#9'L'+
                                             tostr(pai_labeled386(hp)^.lab));

                 { we ignore there comments, because some GNU AS can't handle this }
                 ait_comment : ;
                 ait_symbol : begin
                                 if pai_symbol(hp)^.is_global then
                                   begin
                                      outfile^.write_string('.globl ');
                                      outfile^.write_pchar(pai_symbol(hp)^.name);
                                      outfile^.write_string(target_info.newline);
                                   end;
                                 outfile^.write_pchar(pai_symbol(hp)^.name);
                                 outfile^.write_string(':');
                              end;
                 { writes an instruction, highly table driven }
                 ait_instruction : begin
                                      s:=#9+op2str[pai386(hp)^.operator]+opsize2str[pai386(hp)^.size];
                                      if pai386(hp)^.op1t<>top_none then
                                        begin
                                           { call and jmp need an extra handling                          }
                                           { this code is only callded if jmp isn't a labeled instruction }
                                           if pai386(hp)^.operator in [A_CALL,A_JMP] then
                                             begin
                                                s:=s+#9+getopstr_jmp(pai386(hp)^.op1t,pai386(hp)^.op1);
                                             end
                                           else
                                             begin
                                                s:=s+#9+getopstr(pai386(hp)^.op1t,pai386(hp)^.op1);
                                                if pai386(hp)^.op2t<>top_none then
                                                  begin
                                                     s:=s+','+getopstr(pai386(hp)^.op2t,pai386(hp)^.op2);
                                                     if pai386(hp)^.op3t<>top_none then
                                                       begin
                                                          s:=s+','+getopstr(pai386(hp)^.op3t,pai386(hp)^.op3);
                                                       end;
                                                  end;
                                             end;
                                        end;
                                      outfile^.write_string(s);
                                   end;
                 else internalerror(10000);
              end;
              { omit extra new lines }
              if hp^.typ<>ait_external then
                outfile^.write_string(target_info.newline);
              hp:=pai(hp^.next);
           end;
      end;

    procedure asmoutlists_init;

      begin
         datasegment:=new(pi386attasmoutput,init);
         codesegment:=new(pi386attasmoutput,init);
         bsssegment:=new(pi386attasmoutput,init);
         debuglist:=new(pi386attasmoutput,init);
         externals:=new(pi386attasmoutput,init);
      end;

    procedure writeasmlists(outfile : pbufferedfile);

      begin
         externals^.write_to_file(outfile);

         outfile^.write_string('.text');
         outfile^.write_string(target_info.newline);
         outfile^.write_string(#9'.align 4');
         outfile^.write_string(target_info.newline);
         codesegment^.write_to_file(outfile);

         outfile^.write_string('.data');
         outfile^.write_string(target_info.newline);
         outfile^.write_string(#9'.align 4');
         outfile^.write_string(target_info.newline);

         { some AS versions don't like comments, so we write }
         { an ascii string                                    }
         outfile^.write_string(#9'.ascii "compiled by FPKPascal '+version_string+'\0"');
         outfile^.write_string(target_info.newline);
         outfile^.write_string(#9'.ascii "target: '+target_info.target_name+'\0"');
         outfile^.write_string(target_info.newline);

         outfile^.write_string(#9'.align 4');
         outfile^.write_string(target_info.newline);
         datasegment^.write_to_file(outfile);

         { makes problems with old GNU ASes
         outfile^.write_string('.bss');
         outfile^.write_string(target_info.newline);
         outfile^.write_string(#9'.align 4');
         }
         outfile^.write_string(target_info.newline);
         bsssegment^.write_to_file(outfile);
      end;

end.
