{****************************************************************************

                     Copyright (c) 1993,96 by Florian Klaempfl

 ****************************************************************************}
{$I-}
unit symtable;

  interface

    uses
       objects,cobjects,errors,systems,globals,dos,strings,aasm
{$ifdef i386}
       ,i386
{$endif}
{$ifdef GDB}
       ,gdb
{$endif}
       ;


    const
       { possible types of symtables }
       localsymtable = $8000;
       parasymtable = $4000;

       withsymtable = 1;
       staticsymtable = 2;
       globalsymtable = 3;
       unitsymtable = 4;
       objectsymtable = 5;
       recordsymtable = 6;
       macrosymtable = 7;

       { options of subroutines }
       poexceptions = $1;
       povirtualmethod = $2;
       poclearstack = $4;
       poconstructor = $8;
       podestructor = $10;
       pointernproc = $20;
       poexports = $40;
       poiocheck = $80;
       poabstractmethod = $100;
       pointerrupt = $200;
       poinline = $400;

       hasharraysize = 97;

       { last operator which can be overloaded }
       last_overloaded = PLUS;

{$ifdef GDB}
       { for STAB debugging }
       Const GlobalTypeCount : word = 1;
{$endif * GDB *}
       { options for objects and classes }
       oois_abstract = $1;
       oois_class = $2;

    type
       { "forward" pointer }

       pformaldef = ^tformaldef;
       pfiledef = ^tfiledef;
       pobjectdef = ^tobjectdef;
       precdef = ^trecdef;
       parraydef = ^tarraydef;
       ppointerdef = ^tpointerdef;
       pstringdef = ^tstringdef;
       paufzaehldef = ^taufzaehldef;
       pgrunddef = ^tgrunddef;
       pprocdef = ^tprocdef;
       perrordef = ^terrordef;
       psetdef = ^tsetdef;
       psymtable = ^tsymtable;
       pdef = ^tdef;
       pprocvardef = ^tprocvardef;
       pabstractprocdef = ^tabstractprocdef;
       psym = ^tsym;
       plabelsym = ^tlabelsym;

       { base types }
       tbasetyp = (uauto,u8bit,s32bit,s64real,uvoid,bool8bit,uchar,
                      s8bit,s16bit,u16bit,u32bit);


       { possible types for symtable entries }
       tsymtyp = (abstractsym,varsym,typesym,procsym,unitsym,programsym,
                  constsym,aufzaehlsym,typedconstsym,errorsym,syssym,
                  labelsym);

       { this object is the base for all symbol objects }
       tsym = object
          typ : tsymtyp;
          _name : pchar;
          left : psym;
          right : psym;
          speedvalue : longint;
          forwarddef : boolean;
          owner : psymtable;
{$ifdef GDB}
          isstabwritten : boolean;
{$endif * GDB *}
          constructor init(const n : string);
          constructor load;
          destructor done;virtual;
          procedure write;virtual;
          procedure deref;virtual;
          function name : string;
          procedure setname(const s : string);
{$ifdef GDB}
          function stabstring : pchar;virtual;
          procedure concatstabto(asmlist : paasmoutput);virtual;
{$endif * GDB *}
       end;

       tlabelsym = object(tsym)
          number : longint;
          defined : boolean;
          constructor init(const n : string;l : longint);
          destructor done;virtual;
          procedure write;virtual;
       end;

       punitsym = ^tunitsym;

       tunitsym = object(tsym)
          unitsymtable : psymtable;
          constructor init(const n : string;ref : psymtable);
          procedure write;virtual;
{$ifdef GDB}
          procedure concatstabto(asmlist : paasmoutput);virtual;
{$endif * GDB *}
       end;

       pmacrosym = ^tmacrosym;

       tmacrosym = object(tsym)
          defined : boolean;
          buftext : pchar;
          buflen : longint;
          { macros aren't written to PPU files ! }
          constructor init(const n : string);
          destructor done;virtual;
       end;

       perrorsym = ^terrorsym;

       terrorsym = object(tsym)
          constructor init;
       end;

       pprocsym = ^tprocsym;

       tprocsym = object(tsym)
          definition : pprocdef;
          constructor init(const n : string);
          constructor load;
          destructor done;virtual;
          { tests, if all procedures definitions are defined and not }
          { only forward                                             }
          procedure check_forward;
          procedure write;virtual;
          procedure deref;virtual;
          function is_global : boolean;
{$ifdef GDB}
          function stabstring : pchar;virtual;
          procedure concatstabto(asmlist : paasmoutput);virtual;
{$endif * GDB *}
       end;

       ptypesym = ^ttypesym;

       ttypesym = object(tsym)
          definition : pdef;
          constructor init(const n : string;d : pdef);
          constructor load;
          destructor done;virtual;
          procedure write;virtual;
          procedure deref;virtual;
{$ifdef GDB}
          function stabstring : pchar;virtual;
          procedure concatstabto(asmlist : paasmoutput);virtual;
{$endif * GDB *}
       end;

       tvarspez = (vs_value,vs_const,vs_var);

       pvarsym = ^tvarsym;

       tvarsym = object(tsym)
          adresse : longint;
          definition : pdef;
          refs : longint;
          regable : boolean;

          { if reg<>R_NO, then the variable is an register variable }
          reg : tregister;

          { sets the type of access }
          varspez : tvarspez;
          constructor init(const n : string;p : pdef);
          constructor load;
          function getsize : longint;
          procedure write;virtual;
          procedure deref;virtual;
{$ifdef GDB}
          function stabstring : pchar;virtual;
{$endif * GDB *}
       end;

       ptypedconstsym = ^ttypedconstsym;

       ttypedconstsym = object(tsym)
          prefix : pstring;
          definition : pdef;
          constructor init(const n : string;p : pdef);
          constructor load;
          destructor done;virtual;
          procedure write;virtual;
          procedure deref;virtual;
{$ifdef GDB}
          function stabstring : pchar;virtual;
{$endif * GDB *}
       end;

       tconsttype = (constord,conststring,constreal,constbool,constint,
                     constchar);

       pconstsym = ^tconstsym;

       tconstsym = object(tsym)
          definition : pdef;
          consttype : tconsttype;
          value : longint;
          constructor init(const n : string;t : tconsttype;v : longint;def : pdef);
          constructor load;
{$ifdef GDB}
          destructor done;virtual;
{$endif * GDB *}
          procedure deref;virtual;
          procedure write;virtual;
{$ifdef GDB}
          function stabstring : pchar;virtual;
{$endif * GDB *}
       end;

       paufzaehlsym = ^taufzaehlsym;

       taufzaehlsym = object(tsym)
          value : longint;
          definition : paufzaehldef;
{$ifndef GDB}
          constructor init(const n : string;def : paufzaehldef;v : longint);
{$else * GDB *}
          next : paufzaehlsym;
          constructor init(const n : string;def : paufzaehldef;v : longint;anext : paufzaehlsym);
{$endif * GDB *}
          constructor load;
          procedure write;virtual;
          procedure deref;virtual;
{$ifdef GDB}
          procedure concatstabto(asmlist : paasmoutput);virtual;
{$endif * GDB *}
       end;

       pprogramsym = ^tprogramsym;

       tprogramsym = object(tsym)
          constructor init(const n : string);
       end;

       psyssym = ^tsyssym;

       tsyssym = object(tsym)
          number : longint;
          constructor init(const n : string;l : longint);
          procedure write;virtual;
{$ifdef GDB}
          procedure concatstabto(asmlist : paasmoutput);virtual;
{$endif * GDB *}
       end;

       tcallback = procedure(p : psym);

       tsymtablehasharray = array[0..hasharraysize-1] of psym;

       psymtablehasharray = ^tsymtablehasharray;

       tsymtable = object
          name : pstring;
          datasize : longint;
          wurzel : psym;
	  hasharray : psymtablehasharray;
          next : psymtable;

          { only used for parameter symtable to determine the offset relative }
          { to the frame pointer                                              }
          call_offset : longint;

          { this saves all definition to allow a proper clean up }
          wurzeldef : pdef;
          symtabletype : word;

          { each symtable gets a number }
          unitid : word;

          constructor init(t : word);
          constructor load;
          constructor loadasstruct(typ : word);
          destructor done;virtual;
          procedure check_forwards;
          procedure insert(sym : psym);
          function search(const s : stringid) : psym;
          procedure clear;
          procedure registerdef(p : pdef);
          procedure foreach(proc2call : tcallback);
          procedure allsymbolsused;
          procedure write;
{$ifdef GDB}
          procedure concatstabto(asmlist : paasmoutput);
          procedure concatallstabto(asmlist : paasmoutput);
{$endif * GDB *}
          procedure writeasunit;
          procedure writeasstruct;
          function getdefnr(l : word) : pdef;
       end;

       punitsymtable = ^tunitsymtable;

       tunitsymtable = object(tsymtable)
          checksum,maschstart : longint;
          constructor load(const n : string);
       end;


       { defintion contains the informations about a type }
       tdeftype = (abstractdef,arraydef,recorddef,pointerdef,grunddef,
                   stringdef,aufzaehldef,procdef,objectdef,errordef,
                   filedef,formaldef,setdef,procvardef);

       tdef = object
          savesize : longint;
          owner : psymtable;

          { this allows to determine by which type the defintion was generated }
          sym : ptypesym;
          next : pdef;
{$ifdef GDB}
          nextglobal : pdef;
          isstabwritten : boolean;
          isusedinstab : boolean;
          globalnb : word;
{$endif * GDB *}
          number : word;
          deftype : tdeftype;

          function size : longint;virtual;
          constructor init;
          destructor done;virtual;
          procedure write;virtual;
{$ifdef GDB}
          function NumberString : string;
          constructor load;
          function stabstring : pchar;virtual;
          function allstabstring : pchar;
          procedure concatstabto(asmlist : paasmoutput);virtual;
{$endif * GDB *}
          procedure deref;virtual;
       end;

       tfiletyp = (ft_text,ft_typed,ft_untyped);

       tfiledef = object(tdef)
          public
             filetyp : tfiletyp;
             typed_as : pdef;
             constructor init(ft : tfiletyp;tas : pdef);
             constructor load;
             procedure write;virtual;
{$ifdef GDB}
             function stabstring : pchar;virtual;
             procedure concatstabto(asmlist : paasmoutput);virtual;
{$endif * GDB *}
             procedure deref;virtual;
          private
             procedure setsize;
       end;

       tformaldef = object(tdef)
          constructor init;
          constructor load;
          procedure write;virtual;
{$ifdef GDB}
          function stabstring : pchar;virtual;
          procedure concatstabto(asmlist : paasmoutput);virtual;
{$endif * GDB *}
       end;

       terrordef = object(tdef)
          constructor init;
{$ifdef GDB}
          function stabstring : pchar;virtual;
{$endif * GDB *}
       end;

       tpointerdef = object(tdef)
          definition : pdef;
          constructor init(def : pdef);
          constructor load;
          procedure write;virtual;
{$ifdef GDB}
          function stabstring : pchar;virtual;
          procedure concatstabto(asmlist : paasmoutput);virtual;
{$endif * GDB *}
          procedure deref;virtual;
       end;

       tarraydef = object(tdef)
          lowrange : longint;
          highrange : longint;
          rangenr : longint;
          definition : pdef;
          rangedef : pdef;
          function elesize : longint;
          constructor init(l,h : longint;rd : pdef);
          constructor load;
          procedure write;virtual;
{$ifdef GDB}
          function stabstring : pchar;virtual;
          procedure concatstabto(asmlist : paasmoutput);virtual;
{$endif * GDB *}
          procedure deref;virtual;
          function size : longint;virtual;

          { generates the ranges needed by the asm instruction BOUND }
          procedure genrangecheck;
       end;

       trecdef = object(tdef)
          symtable : psymtable;
          constructor init(p : psymtable);
          constructor load;
          destructor done;virtual;
          procedure write;virtual;
{$ifdef GDB}
          function stabstring : pchar;virtual;
          procedure concatstabto(asmlist : paasmoutput);virtual;
{$endif * GDB *}
          procedure deref;virtual;
       end;

       tgrunddef = object(tdef)
          von : longint;
          bis : longint;
          rangenr : longint;
          typ : tbasetyp;
          constructor init(t : tbasetyp;v,b : longint);
          constructor load;
          procedure write;virtual;
{$ifdef GDB}
          function stabstring : pchar;virtual;
{$endif * GDB *}
          procedure setsize;

          { generates the ranges needed by the asm instruction BOUND }
          procedure genrangecheck;
       end;

       pdefcoll = ^tdefcoll;

       tdefcoll = record
          data : pdef;
          next : pdefcoll;
          paratyp : tvarspez;
       end;

       tabstractprocdef = object(tdef)
          { saves a defintion to the return type }
          retdef : pdef;
          { save the procedure options }
          options : word;
          para1 : pdefcoll;
          constructor init;
          constructor load;
          destructor done;virtual;
          procedure concatdef(p : pdef;vsp : tvarspez);
          procedure deref;virtual;
{$ifdef GDB}
          function stabstring : pchar;virtual;
          procedure concatstabto(asmlist : paasmoutput);virtual;
{$endif * GDB *}
          procedure write;virtual;
       end;

       tprocvardef = object(tabstractprocdef)
          constructor init;
          constructor load;
          procedure write;virtual;
{$ifdef GDB}
          function stabstring : pchar;virtual;
{$endif * GDB *}
       end;

       tprocdef = object(tabstractprocdef)

          extnumber : longint;
          nextoverloaded : pprocdef;
          { pointer to the local symbol table }
          localst : psymtable;
          { pointer to the parameter symbol table }
          parast : psymtable;


          _class : pobjectdef;
          _mangledname : pchar;

          { it's a tree, but this not easy to handle }
          { with the interfaces of units             }
          code : pointer;

          { true, if the procedure is only declared }
          { (forward procedure) }
          forwarddef : boolean;

          { set which contains the modified registers }
{$ifdef i386}
          usedregisters : byte;
{$endif}
          constructor init;
          destructor done;virtual;
          constructor load;
          procedure write;virtual;
{$ifdef GDB}
          function stabstring : pchar;virtual;
          procedure concatstabto(asmlist : paasmoutput);virtual;
{$endif * GDB *}
          procedure deref;virtual;
          function mangledname : string;
          procedure setmangledname(const s : string);
       end;

       tstringdef = object(tdef)
          len : byte;
          constructor init(l : byte);
          constructor load;
          procedure write;virtual;
{$ifdef GDB}
          function stabstring : pchar;virtual;
          procedure concatstabto(asmlist : paasmoutput);virtual;
{$endif * GDB *}
       end;

       taufzaehldef = object(tdef)
          max : longint;
{$ifdef GDB}
          first : paufzaehlsym;
{$endif * GDB *}
          constructor init;
          constructor load;
{$ifdef GDB}
          destructor done;virtual;
{$endif * GDB *}
          procedure write;virtual;
{$ifdef GDB}
          function stabstring : pchar;virtual;
{$endif * GDB *}
       end;

       tobjectdef = object(tdef)
          childof : pobjectdef;
          name : pstring;
{          privatesyms : psymtable;
          protectedsyms : psymtable; }
          publicsyms : psymtable;
          options : word;
          constructor init(const n : string;c : pobjectdef);
          destructor done;virtual;
          function isrelated(d : pobjectdef) : boolean;
          function size : longint;virtual;
          constructor load;
          procedure write;virtual;
{$ifdef GDB}
          function stabstring : pchar;virtual;
          procedure concatstabto(asmlist : paasmoutput);virtual;
{$endif * GDB *}
          procedure deref;virtual;
       end;

       tsettyp = (normset);

       tsetdef = object(tdef)
          setof : pdef;
          settyp : tsettyp;
          constructor init(s : pdef;high : longint);
          constructor load;
          procedure write;virtual;
{$ifdef GDB}
          function stabstring : pchar;virtual;
          procedure concatstabto(asmlist : paasmoutput);virtual;
{$endif * GDB *}
          procedure deref;virtual;
       end;

       punitnamerec = ^tunitnamerec;

       tunitnamerec = record
          unitname : pstring;
          next : punitnamerec;
       end;

    { inits the symbol table administration }
    procedure init_symtable;
    procedure done_symtable;

    procedure getsym(const s : stringid;notfounderror : boolean);
    procedure getsymonlyin(p : psymtable;const s : stringid);

    { writes an unit with the given name, }
    { returns the filesize                }
    function writeunitas(const s : string;unitsymtable : psymtable) : longint;

    { deletes a symbol table from the symbol table stack }
    procedure dellexlevel;

    { saves a forward pointer defintion .... }
    procedure save_forward(ppd : ppointerdef;typesym : ptypesym);

    { .... resolves this forward definitions }
    procedure resolve_forwards;

    var
       registerdef : boolean;      { true, wenn Definitionen           }
                                   { registriert werden sollen         }

       symtablestack : psymtable;  { Wurzel der verketteten Liste von  }
                                   { Symboltabellen                    }

       srsym : psym;               { enthlt das Ergebnis der letzten  }
       srsymtable : psymtable;     { Suche nach einem Symbol           }

       forwardsallowed : boolean;  { true, wenn Pointertypen "forward" }
                                   { eingefgt werden drfen           }

       constsymtable : psymtable;  { Symboltabelle in die die          }
                                   { Konstanten von z.B. Aufzhlungs-  }
                                   { typen eingefgt werden            }

       voiddef : pgrunddef;        { Zeiger auf eine void-Definition   }
                                   { wird von quelltext initialisiert  }
                                   { (ist resulttype einer Procedure)  }
       voidpointerdef : ppointerdef;
                                   { Zeiger auf "void"-Pointerdef      }

       s32bitdef : pgrunddef;      { Zeiger fr resulttype von         }
                                   { intconstn                         }

       u8bitdef : pgrunddef;       { Pointer auf 8-Bit unsigned        }
       u16bitdef : pgrunddef;      { Pointer auf 8-Bit unsigned        }

       cs64realdef : pgrunddef;    { Zeiger fr resulttype von         }
                                   { realconstn                        }

       cstringdef : pstringdef;    { Zeiger fr resulttype von         }
                                   { stringconstn                      }

       cchardef : pgrunddef;       { Zeiger fr resulttype von         }
                                   { charconstn                        }

{$ifdef GDB}
       firstglobaldef, lastglobaldef : pdef;
       {uses for stabs}

{$endif * GDB *}
       class_tobject : pobjectdef; { pointer to the anchestor of all   }
                                   { clases                            }

       booldef : pgrunddef;        { pointer to boolean type           }

       aktprocsym : pprocsym;      { Zeiger auf den Symboltablellen-   }
                                   { eintrag der momentan geparseten   }
                                   { procedure                         }

       procprefix : string;        { eindeutige Namen bei geschachtel- }
                                   { ten Unterprogrammen erzeugen      }

       lexlevel : word;            { Stufen von verschachtelten        }
                                   { Unterprogrammen                   }

       macros : psymtable;         { Zeiger auf die Symboltabelle mit  }
                                   { Makros                            }
       usedunits : pstringcontainer;
                                   { enthaelt die Namen aller zu       }
                                   { initialisierenden Units           }

       read_member : boolean;      { true, wenn Members aus einer PPU-  }
                                   { Datei gelesen werden, d.h. ein     }
                                   { varsym seine Adresse einlesen soll }

       generrorsym : psym;         { Jokersymbol, wenn das richtige    }
                                   { Symbol nicht gefunden wird        }

       generrordef : pdef;         { Jokersymbol fr eine fehlerhafte  }
                                   { Typdefinition                     }

       {
         !!!! overloaded_operators : array[PLUS..last_overloaded] of pprocsym;
       }
       overloaded_operators : array[0..0] of pprocsym;


    const
{$ifdef tp}
       maxunitsize = 65000;
{$else}
       maxunitsize = 1024*1024*16;
{$endif}

                                           {                 Compilerversion }
                                           {             Format      |       }
                                           { Kennung      |          |       }
                                           {  |           |          |       }
                                           { /-------\   /-------\  /---\    }
{$IfDef GDB  changes the unit file format to keep some information
otherwise lost }
       unitheader : array[0..19] of char = ('P','P','U','5','1','1',#0,#6,
{$Else GDB}
       unitheader : array[0..19] of char = ('P','P','U','0','1','1',#0,#6,
{$EndIf GDB}
                                            #0,#255,#0,#0,#0,#0,#255,#255,
                                           { |   |  \---------/ \-------/    }
                                           { |   |    |             |        }
                                           { |   |    Checksumme    |        }
                                           { |   \--momentan unbenutzt       }
                                           { Zielbetriebssystem              }
                                            #0,#0,#0,#0);
                                           {\---------/                      }
                                           {  |                              }
                                           {  Start der Maschinensprache     }

    type
       trbuffer = array[0..maxunitsize] of byte;
       prbuffer = ^trbuffer;

    var
       rbuffer : prbuffer;
       bufferl,bufferp : word;

    type
       punit = ^tunit;

       tunit = object
          { contains the global symtable of the unit     }
          { p=nil means that the unit must be recompiled }
          symtab : punitsymtable;

          { path to the source file }
          path : pstring;
          name : pstring;
          sourcename : pstring;
          sourceext : pstring;
          next : punit;
          constructor init(const p,n : string;anext : punit);
       end;

    function searchcompile : punit;
    function readunit(const s : string) : punitsymtable;
    function searchunit(const s : string) : punit;

    var
       units : punit;

{$ifdef GDB}
    procedure concatdefstabto(asmlist : paasmoutput);
    function typeglobalnumber(const s : string) : string;
{$endif * GDB *}

implementation

    var
       aktrecordsymtable : psymtable; { zeigt auf die Symboltabelle des }
                                      { Records, das momentan aus einer }
                                      { PPU-Datei gelesen wird          }

    const
       ibloadunit = 1;
       ibgrunddef = 2;
       ibpointerdef = 3;
       ibtypesym = 4;
       ibarraydef = 5;
       ibprocdef = 6;
       ibprocsym = 7;
       iblinkofile = 8;
       ibstringdef = 9;
       ibvarsym = 10;
       ibconstsym = 11;
       ibinitunit = 12;
       ibaufzaehlsym = 13;
       ibtypedconstsym = 14;
       ibrecorddef = 15;
       ibfiledef = 16;
       ibformaldef = 17;
       ibobjectdef = 18;
       ibaufzaehldef = 19;
       ibsetdef = 20;
       ibprocvardef = 21;
       ibend = 255;

    const
{$ifdef tp}
       buffersize = 10000;
{$else}
       buffersize = 1024*1024;
{$endif}

    type
       tubuffer = array[0..buffersize-1] of byte;
       txbuffer = array[0..buffersize div 4] of longint;

       pxbuffer = ^txbuffer;

    var
       unitfile : file;
       buffer : ^tubuffer;
       checksum : longint;

    procedure writebuffer;

      var
         i : longint;

      begin
         for i:=0 to ((bufferp-1) div 4)-1 do
           checksum:=checksum xor pxbuffer(buffer)^[i];
         blockwrite(unitfile,buffer^,bufferp);
         bufferp:=0;
      end;

    procedure writebyte(b : byte);

      begin
         if bufferp>=buffersize then
           writebuffer;
         buffer^[bufferp]:=b;
         inc(bufferp);
      end;

    procedure writelong(l : longint);

      begin
         if bufferp>=buffersize-3 then
           writebuffer;
         move(l,buffer^[bufferp],4);
         inc(bufferp,4);
      end;

    procedure writedouble(d : double);

      begin
         if bufferp>=buffersize-7 then
           writebuffer;
         move(d,buffer^[bufferp],8);
         inc(bufferp,8);
      end;

    procedure writeword(w : word);

      begin
         if bufferp>=buffersize-1 then
           writebuffer;
         move(w,buffer^[bufferp],2);
         inc(bufferp,2);
      end;

    procedure writestring(const s : string);

      begin
         if bufferp>=buffersize-length(s) then
           writebuffer;
         move(s[0],buffer^[bufferp],length(s)+1);
         inc(bufferp,length(s)+1);
      end;

    procedure writedefref(p : pdef);

      begin
         if p=nil then
           writelong($ffffffff)
         else
           begin
              if (p^.owner^.symtabletype=recordsymtable) or
                 (p^.owner^.symtabletype=objectsymtable) then
                writeword($ffff)
              else writeword(p^.owner^.unitid);
              writeword(p^.number);
           end;
      end;

    function writeunitas(const s : string;unitsymtable : psymtable) : longint;

      var
         size : longint;
         maschstart : longint;

      begin
         assign(unitfile,s);
         rewrite(unitfile,1);
         if ioresult<>0 then
           fatalerror(cannot_write_unitfile);
         getmem(buffer,buffersize);
         { char(...) makes a problem with FPKPascal }
         unitheader[8]:=char(byte(target_info.target));
         move(unitheader[0],buffer^,sizeof(unitheader));
         bufferp:=sizeof(unitheader);
         checksum:=0;
         unitsymtable^.writeasunit;

         { now write the rest in the buffer }
         writebuffer;

{$ifdef aout}
         maschstart:=bufferp;
{$endif}

         seek(unitfile,10);
         blockwrite(unitfile,checksum,4);

{$ifdef aout}
         seek(unitfile,16);
         blockwrite(unitfile,maschstart,4);
{$endif}

         size:=filesize(unitfile);
         close(unitfile);
         freemem(buffer,buffersize);
         writeunitas:=size;
      end;

    type
       ploaded_units = ^tloaded_units;

       tloaded_units = record
          name : pstring;
          unitpos : psymtable;
          next : ploaded_units;
       end;

    var
       loaded_units : ploaded_units;

    function searchsymtable(name : stringid) : psymtable;

      var
         hs : ploaded_units;

      begin
          hs:=loaded_units;
          while (assigned(hs)) do
            if (hs^.name^=name) then
              begin
                 searchsymtable:=hs^.unitpos;
                 exit;
              end
            else
              hs:=hs^.next;
          searchsymtable:=nil;
       end;

    const
{$ifdef tp}
       maxsymtables = 255;
{$else}
       maxsymtables = 1024;
{$endif}

    type
       tsymtablemap = array[0..maxsymtables] of psymtable;
       psymtablemap = ^tsymtablemap;

    var
       aktsymtablemap : psymtablemap;

    function readbyte : byte;

      begin
         if bufferp>bufferl-1 then
           fatalerror(error_reading_unit);
         inc(bufferp);
         readbyte:=rbuffer^[bufferp-1];
      end;

    function readword : word;

      var
         w : word;

      begin
         if bufferp>bufferl-2 then
           fatalerror(error_reading_unit);
         move(rbuffer^[bufferp],w,2);
         inc(bufferp,2);
         readword:=w;
      end;

    function readlong : longint;

      var
         l : longint;

      begin
         if bufferp>bufferl-4 then
           fatalerror(error_reading_unit);
         move(rbuffer^[bufferp],l,4);
         inc(bufferp,4);
         readlong:=l;
      end;

    function readdouble : double;

      var
         d : double;

      begin
         if bufferp>bufferl-8 then
           fatalerror(error_reading_unit);
         move(rbuffer^[bufferp],d,8);
         inc(bufferp,8);
         readdouble:=d;
      end;

    function readstring : string;

      var
         s : string;

      begin
         s[0]:=char(readbyte);
         if bufferp>bufferl-ord(s[0]) then
           fatalerror(error_reading_unit);
         move(rbuffer^[bufferp],s[1],ord(s[0]));
         inc(bufferp,ord(s[0]));
         readstring:=s;
      end;

    function readdefref : pdef;

      var
         hd : pdef;

      begin
         longint(hd):=readword;
         longint(hd):=longint(hd) or (longint(readword) shl 16);
         readdefref:=hd;
      end;

    procedure resolvedef(var d : pdef);

      begin
         if longint(d)=$ffffffff then
           d:=nil
         else
           begin
              if (longint(d) and $ffff)=$ffff then
                d:=aktrecordsymtable^.getdefnr(longint(d) shr 16)
              else
                d:=aktsymtablemap^[longint(d) and $ffff]^.getdefnr(longint(d) shr 16);
           end;
      end;

{$I+}
    procedure getsym(const s : stringid;notfounderror : boolean);

      begin
         srsymtable:=symtablestack;
         while assigned(srsymtable) do
           begin
              srsym:=srsymtable^.search(s);
              if assigned(srsym) then exit
               else srsymtable:=srsymtable^.next;
           end;
         if forwardsallowed then
           begin
              srsymtable:=symtablestack;
              srsym:=new(ptypesym,init(s,nil));
              srsym^.forwarddef:=true;
              srsymtable^.insert(srsym);
           end
         else if notfounderror then
           begin
              exterror:=strpnew(s);
              error(id_not_found);
              srsym:=generrorsym;
           end
         else srsym:=nil;
      end;

    procedure getsymonlyin(p : psymtable;const s : stringid);

      begin
         srsymtable:=p;
         srsym:=srsymtable^.search(s);
         if assigned(srsym) then exit
         else fatalerror(id_not_found);
      end;

    procedure dellexlevel;

      var
         p : psymtable;

      begin
         p:=symtablestack;
         symtablestack:=p^.next;

         { symbol tables of unit interfaces are never disposed }
         { this is handle by the unit unitm                    }
         if (p^.symtabletype<>unitsymtable) and
           (p^.symtabletype<>globalsymtable) then
           dispose(p,done);
      end;

    constructor tprocsym.init(const n : string);

      begin
         tsym.init(n);
         typ:=procsym;
         definition:=nil;
      end;

    constructor tprocsym.load;

      begin
         tsym.load;
         typ:=procsym;
         definition:=pprocdef(readdefref);
      end;

    destructor tprocsym.done;

      begin
         check_forward;
         tsym.done;
      end;

    procedure tprocsym.check_forward;

      var
         pd : pprocdef;

      begin
         pd:=definition;
         while assigned(pd) do
           begin
              if pd^.forwarddef then
                begin
                   exterror:=strpnew(name);
                   error(forward_not_resolved);
                end;
              pd:=pd^.nextoverloaded;
           end;
      end;

    procedure tprocsym.deref;

      begin
         resolvedef(pdef(definition));
      end;

    constructor tprogramsym.init(const n : string);

      begin
         tsym.init(n);
         typ:=programsym;
      end;

    constructor tsymtable.init(t : word);

      var
         w : word;

      begin
         symtabletype:=t;
         wurzel:=nil;
         next:=nil;
         name:=nil;
{$ifdef GDB}
         call_offset:=0;
{$endif * GDB *}
         if symtabletype=objectsymtable then
           datasize:=4
         else
           datasize:=0;
         wurzeldef:=nil;

{$ifdef GDB}
         unitid := 0;
{$endif * GDB *}
         case symtabletype of
         { only symtable which become big, get a hash array }
            globalsymtable,staticsymtable,unitsymtable :
              begin
                 new(hasharray);
                 for w:=0 to hasharraysize-1 do
                   hasharray^[w]:=nil;
              end;
            else hasharray:=nil;
         end;
      end;

    constructor tsymtable.load;

      var
         hp : pdef;
         b : byte;
         counter : word;
         hs : punitsymtable;
         map : psymtablemap;
         nextmapentry : word;
         sym : psym;
         unittoload : stringid;
         unitchecksum : longint;
         hr : tunitnamerec;
         unitfilename : pathstr;
{$ifndef GDB}

{$else * GDB *}
         {to know where the unit is }
         dir : string;
{$endif * GDB *}
      begin
         new(map);
         map^[0]:=@self;
         nextmapentry:=1;

         symtabletype:=unitsymtable;

{$ifdef GDB}
         call_offset:=0;
{$endif * GDB *}

         { reset hash array }
         new(hasharray);
         for counter:=0 to hasharraysize-1 do
            hasharray^[counter]:=nil;

         datasize:=0;
         wurzel:=nil;
         next:=nil;
         wurzeldef:=nil;

{$ifdef GDB}
         unitid:=0;
{$endif * GDB *}

         { read the definitions }
         counter:=0;
         repeat
           b:=readbyte;
           case b of
              ibloadunit : begin
                              unittoload:=readstring;
                              unitchecksum:=readlong;
                              hs:=readunit(unittoload);

                              { stimmt die Checksumme in der geladenen Unit }
                              { mit der Checksumme in der momentan zu       }
                              { ladenden Unit berein ?                     }
                              if unitchecksum<>hs^.checksum then
                                begin
                                   { yes, then print a warning }
                                   exterror:=strpnew(name^);
                                   warning(ill_unit_version);
                                end;
                              map^[nextmapentry]:=hs;
                              inc(nextmapentry);
                              if nextmapentry>maxsymtables then
                                fatalerror(too_much_units);
                           end;
              ibpointerdef : hp:=new(ppointerdef,load);
              ibarraydef : hp:=new(parraydef,load);
              ibgrunddef : hp:=new(pgrunddef,load);
              ibprocdef : hp:=new(pprocdef,load);
              ibstringdef : hp:=new(pstringdef,load);
              ibrecorddef : hp:=new(precdef,load);
              ibobjectdef : hp:=new(pobjectdef,load);
              ibfiledef : hp:=new(pfiledef,load);
              ibformaldef : hp:=new(pformaldef,load);
              ibaufzaehldef : hp:=new(paufzaehldef,load);
              ibinitunit : usedunits^.insert(readstring);
              iblinkofile : linkofiles.insert(readstring);
              ibsetdef : hp:=new(psetdef,load);
              ibprocvardef : hp:=new(pprocvardef,load);
              ibend : break;
              else fatalerror(malformed_unit);
           end;

           if (b<>ibloadunit) and (b<>ibinitunit) and (b<>iblinkofile) then
             begin
                { each definition get a number }
                hp^.number:=counter;
                inc(counter);

                hp^.next:=wurzeldef;
                wurzeldef:=hp;
             end;
         until false;

         { solve the references of the symbols }
         aktsymtablemap:=map;
         hp:=wurzeldef;

         { for each definition }
         while assigned(hp) do
           begin
              hp^.deref;

              { insert also the owner }
              hp^.owner:=@self;

              hp:=hp^.next;
           end;

         { read the symbols }
         repeat
           b:=readbyte;
           case b of
              ibtypesym : sym:=new(ptypesym,load);
              ibprocsym : sym:=new(pprocsym,load);
              ibconstsym : sym:=new(pconstsym,load);
              ibvarsym : sym:=new(pvarsym,load);
              ibaufzaehlsym : sym:=new(paufzaehlsym,load);
              ibtypedconstsym : sym:=new(ptypedconstsym,load);
              ibend : break;
              else fatalerror(malformed_unit);
           end;
           sym^.deref;
           insert(sym);
         until false;

         { the map about the defintions is never needed }
         dispose(map);
      end;

    constructor tunitsymtable.load(const n : string);

      var
         hp : ploaded_units;

      begin
         name:=stringdup(upper(n));
         new(hp);
         hp^.name:=name;
         hp^.unitpos:=@self;
         hp^.next:=loaded_units;
         loaded_units:=hp;
         inherited load;
      end;

    constructor tsymtable.loadasstruct(typ : word);

      var
         hp : pdef;
         b : byte;
         counter : word;
         sym : psym;

      begin
         symtabletype:=typ;
         hasharray:=nil;
         aktrecordsymtable:=@self;
         name:=nil;
         if symtabletype=objectsymtable then
           datasize:=4
         else
           datasize:=0;
{$ifdef GDB}
         call_offset := 0;
{$endif * GDB *}
         wurzel:=nil;
         next:=nil;
         wurzeldef:=nil;
{$ifdef GDB}
         unitid := 0;
{$endif * GDB *}

         { read definitions }
         counter:=0;
         repeat
           b:=readbyte;
           case b of
              ibpointerdef : hp:=new(ppointerdef,load);
              ibarraydef : hp:=new(parraydef,load);
              ibgrunddef : hp:=new(pgrunddef,load);
              ibprocdef : hp:=new(pprocdef,load);
              ibstringdef : hp:=new(pstringdef,load);
              ibrecorddef : hp:=new(precdef,load);
              ibobjectdef : hp:=new(pobjectdef,load);
              ibaufzaehldef : hp:=new(paufzaehldef,load);
              ibsetdef : hp:=new(psetdef,load);
              ibprocvardef : hp:=new(pprocvardef,load);
              ibfiledef : hp:=new(pfiledef,load);
              ibformaldef : hp:=new(pformaldef,load);
              ibend : break;
              else fatalerror(malformed_unit);
           end;

           { Numerieren }
           hp^.number:=counter;
           inc(counter);
           hp^.next:=wurzeldef;
           wurzeldef:=hp;
         until false;
         { dereferenziert wird erst in trecdef^.deref }
         { nun Symbole einlesen }
         repeat
           b:=readbyte;
           case b of
              ibtypesym : sym:=new(ptypesym,load);
              ibprocsym : sym:=new(pprocsym,load);
              ibconstsym : sym:=new(pconstsym,load);
              ibvarsym : sym:=new(pvarsym,load);
              ibaufzaehlsym : sym:=new(paufzaehlsym,load);
              ibtypedconstsym : sym:=new(ptypedconstsym,load);
              ibend : break;
              else fatalerror(malformed_unit);
           end;
           insert(sym);
         until false;
      {$IfDef GDB}
      call_offset := readlong;
      {$EndIf GDB}
      end;

    destructor tsymtable.done;

      var
         hp : pdef;
{$ifndef GDB}

{$else * GDB *}
         last : pdef;
{$endif * GDB *}
      begin
         { erst die Eintrge loeschen, da procsym's noch ihre Definitionen }
         { auf unaufgelste "forwards" ueberpruefen                        }
         clear;
{$ifdef GDB}
         stringdispose(name);
{$endif * GDB *}
         hp:=wurzeldef;
{$ifdef GDB}
         last := Nil;
{$endif * GDB *}
         while assigned(hp) do
           begin
{$ifdef GDB}
              if hp^.owner = @Self then
                begin
                if assigned(last) then last^.next := hp^.next;
{$endif * GDB *}
              wurzeldef:=hp^.next;
              dispose(hp,done);
{$ifdef GDB}
                end else
                begin
                last := hp;
                wurzeldef:=hp^.next;
                end;
{$endif * GDB *}
              hp:=wurzeldef;
           end;

      end;

    procedure check_procsym_forward(sym : psym);far;

      begin
         if sym^.typ=procsym then
           pprocsym(sym)^.check_forward;
      end;

    { checks, if all procsyms are defined }
    procedure tsymtable.check_forwards;

      begin
{$ifdef tp}
         foreach(check_procsym_forward);
{$else}
         foreach(@check_procsym_forward);
{$endif}
      end;

    function tsymtable.getdefnr(l : word) : pdef;

      var
         hp : pdef;

      begin
         hp:=wurzeldef;
         while (assigned(hp)) and (hp^.number<>l) do
           hp:=hp^.next;
         getdefnr:=hp;
      end;

    procedure tsymtable.registerdef(p : pdef);

      begin
         p^.next:=wurzeldef;
         wurzeldef:=p;
         p^.owner:=@self;
      end;

    procedure tsymtable.clear;

      var
         w : integer;

      begin
         { remove all entry from a symbol table }
         if assigned(wurzel) then
           dispose(wurzel,done);
         if assigned(hasharray) then
           begin
              for w:=0 to hasharraysize-1 do
                if assigned(hasharray^[w]) then
                  dispose(hasharray^[w],done);
              dispose(hasharray);
           end;
      end;

    function getspeedvalue(const s : string) : longint;

      var
         l : longint;
         w : word;

      begin
         l:=0;
         for w:=1 to length(s) do
           l:=l+ord(s[w]);
         getspeedvalue:=l;
      end;

    procedure tsymtable.insert(sym : psym);

      procedure _insert(var osym : psym);

        begin
           if osym=nil then osym:=sym
           { speedvalue is used, to allow a fast insert }
           else if osym^.speedvalue>sym^.speedvalue then _insert(osym^.right)
           else if osym^.speedvalue<sym^.speedvalue then _insert(osym^.left)
           else
             begin
                if osym^.name>sym^.name then _insert(osym^.right)
                else if osym^.name<sym^.name then _insert(osym^.left)
                else
                  begin
                     if (osym^.typ=typesym) and osym^.forwarddef then
                       begin
                          if (sym^.typ<>typesym) then fatalerror(id_already_type);
                          if (ptypesym(sym)^.definition^.deftype<>recorddef) and
                             (ptypesym(sym)^.definition^.deftype<>objectdef) then
                             fatalerror(type_must_be_rec_or_class);
                          ptypesym(osym)^.definition:=ptypesym(sym)^.definition;
                          osym^.forwarddef:=false;
{$ifndef GDB}
                          dispose(sym);
{$else * GDB *}
                          if ptypesym(osym)^.definition^.sym = ptypesym(sym) then
                            ptypesym(osym)^.definition^.sym := ptypesym(osym);
                          dispose(sym,done);
{$endif * GDB *}
                       end
                     else
                       begin
                          exterror:=strpnew(sym^.name);
                          error(dupid);
                       end;
                  end;
             end;
      end;

      var
         l : longint;
         hp : psymtable;
         hsym : psym;

      begin
         { bei Symbolen fr Variablen die Adresse eintragen, }
         { und Gre der Symboltabellendaten berechnen       }
{$ifdef GDB}
         sym^.owner:=@self;
{$endif * GDB *}
         if (sym^.typ=varsym) and not(read_member) then
           begin
              { check for instance of an abstract object or class }
              if (pvarsym(sym)^.definition^.deftype=objectdef) and
                ((pobjectdef(pvarsym(sym)^.definition)^.options and oois_abstract)<>0) then
                error(no_instance_of_abstract_object);

              { bei einer lokalen Symboltabelle erst! erhhen, da der }
              { Wert in codegen.secondload dann mit minus verwendet   }
              { wird                                                  }
              l:=pvarsym(sym)^.getsize;
              if (symtabletype and $8000)=localsymtable then
                begin
                   inc(datasize,l);
                   if (l>=4) and ((datasize and 3)<>0) then
                     inc(datasize,4-(datasize and 3))
                   else if (l>=2) and ((datasize and 1)<>0) then
                     inc(datasize,2-(datasize and 1));

                   pvarsym(sym)^.adresse:=datasize;
                end
              else if (symtabletype and $3fff)=staticsymtable then
                begin
{$ifdef GDB}
                   if cs_debuginfo in aktswitches then
                     begin
                        sym^.concatstabto(bsssegment);
                     end;
{$endif * GDB *}
                   bsssegment^.concat(new(pai_datablock,init('_'+sym^.name,l)));
                   inc(datasize,l);

                   { this symbol can't be loaded to a register }
                   pvarsym(sym)^.regable:=false;
                end
              else if (symtabletype and $3fff)=globalsymtable then
                begin
{$ifdef GDB}
                   if cs_debuginfo in aktswitches then
                     begin
                        sym^.concatstabto(bsssegment);
                        { this has to be added so that the debugger knows where to find
                          the global variable
                        }
                        bsssegment^.concat(new(pai_symbol,init('_'+sym^.name)));
                     end;
{$endif * GDB *}
                   bsssegment^.concat(new(pai_datablock,init_global(
                     'U_'+name^+'_'+sym^.name,l)));
                   inc(datasize,l);

                   { this symbol can't be loaded to a register }
                   pvarsym(sym)^.regable:=false;
                end
              else if ((symtabletype and $3fff)=recordsymtable) or
                      ((symtabletype and $3fff)=objectsymtable) then
                begin
                   { align record and object fields }
                   if aktpackrecords=2 then
                     begin
                        { align to word }
                        if (l>=2) and ((datasize and 1)<>0) then
                          inc(datasize);
                     end
                   else if aktpackrecords=4 then
                     begin
                        { align to dword }
                        if (l>=3) and ((datasize and 3)<>0) then
                          inc(datasize,4-(datasize and 3))
                        { or word }
                        else if (l=2) and ((datasize and 1)<>0) then
                          inc(datasize)
                     end;
                   pvarsym(sym)^.adresse:=datasize;
                   inc(datasize,l);

                   { this symbol can't be loaded to a register }
                   pvarsym(sym)^.regable:=false;
                end
              else if (symtabletype and $4000)=parasymtable then
                begin
                   pvarsym(sym)^.adresse:=datasize;

                   { intel processors don't know a byte push, }
                   { so is always a word pushed               }
                   if l=1 then
                     l:=2;
                   inc(datasize,l);
                end
              else
                begin
                   if (l>=4) and ((datasize and 3)<>0) then
                     inc(datasize,4-(datasize and 3))
                   else if (l>=2) and ((datasize and 1)<>0) then
                     inc(datasize,2-(datasize and 1));
                   pvarsym(sym)^.adresse:=datasize;
                   inc(datasize,l);
                end;
           end
         else if sym^.typ=typedconstsym then
             begin
                if (symtabletype and $3fff)=globalsymtable then
                  begin
{$ifdef GDB}
                     if cs_debuginfo in aktswitches then
                       sym^.concatstabto(datasegment);
{$endif * GDB *}
                     datasegment^.concat(new(pai_symbol,init_global('TC_'+
                       ptypedconstsym(sym)^.prefix^+'_'+sym^.name)));
                  end
                else if not((symtabletype and $3fff)=unitsymtable) then
                  begin
{$ifdef GDB}
                     if cs_debuginfo in aktswitches then
                       sym^.concatstabto(datasegment);
{$endif * GDB *}
                     datasegment^.concat(new(pai_symbol,init('TC_'+
                       ptypedconstsym(sym)^.prefix^+'_'+sym^.name)));
                  end;
             end;
         if (symtabletype=staticsymtable) or
            (symtabletype=globalsymtable) then
           begin
              hp:=symtablestack;
              while assigned(hp) do
                begin
                   if (hp^.symtabletype and $3fff=staticsymtable) or
                      (hp^.symtabletype and $3fff=globalsymtable) then
                        begin
                           hsym:=hp^.search(sym^.name);
                           if (assigned(hsym)) and
                              not(hsym^.forwarddef) then
                              begin
                                 exterror:=strpnew(sym^.name);
                                 error(dupid);
                              end;
                        end;
                      hp:=hp^.next;
                end;
           end;
         sym^.speedvalue:=getspeedvalue(sym^.name);
         if assigned(hasharray) then
           _insert(hasharray^[sym^.speedvalue mod hasharraysize])
         else
           _insert(wurzel);
      end;

    procedure varsymbolused(p : psym);far;

      begin
         if p^.typ=varsym then
           { unused symbol should be reported only if no }
           { error is reported                           }
           if (pvarsym(p)^.refs=0) and (errorcount=0) then
             begin
                exterror:=strpnew(p^.name);
                warning(symbol_not_used);
             end;
      end;

    procedure tsymtable.allsymbolsused;

      begin
{$ifdef tp}
         foreach(varsymbolused);
{$else}
         foreach(@varsymbolused);
{$endif}
      end;

{$ifdef GDB}

    var
       l : paasmoutput;

    procedure concatstab(p : psym);far;

      begin
         if p^.typ<>procsym then
           p^.concatstabto(l);
      end;

    procedure tsymtable.concatstabto(asmlist : paasmoutput);

      begin
         l:=asmlist;
{$ifdef tp}
         foreach(concatstab);
{$else}
         foreach(@concatstab);
{$endif}
      end;

    procedure forcestabto(asmlist : paasmoutput;pd : pdef);

      begin
         if not pd^.isstabwritten then
           begin
              pd^.isusedinstab:=true;
              pd^.concatstabto(asmlist);
           end;
      end;

   procedure concatdefstabto(asmlist : paasmoutput);

     var
        def : pdef;

     begin
        def:=firstglobaldef;
        forcestabto(asmlist,s32bitdef);
        forcestabto(asmlist,cchardef);
        while assigned(def) do
          begin
             { write all named types and only the used unnamed ones !}
             if assigned(def^.sym) then forcestabto(asmlist,def)
               else def^.concatstabto(asmlist);
             def:=def^.nextglobal;
          end;
     end;

    procedure tsymtable.concatallstabto(asmlist : paasmoutput);

      var
         st : psymtable;

      begin
         st:=@self;
         concatdefstabto(asmlist);
         while assigned(st) do
           begin
              st^.concatstabto(asmlist);
              st:=st^.next;
           end;
      end;

{$endif * GDB *}
    function tsymtable.search(const s : stringid) : psym;

      var
         hp : psym;
         w : word;
         speedvalue : longint;

      begin
         speedvalue:=getspeedvalue(s);
         if assigned(hasharray) then
           hp:=hasharray^[speedvalue mod hasharraysize]
         else
           hp:=wurzel;
         while assigned(hp) do
           begin
              if speedvalue>hp^.speedvalue then hp:=hp^.left
              else if speedvalue<hp^.speedvalue then hp:=hp^.right
              else
                begin
                   if hp^.name=s then
                     begin
                        search:=hp;
                        exit;
                     end
                  else if s>hp^.name then hp:=hp^.left
                  else hp:=hp^.right;
                end;
           end;
         search:=nil;
      end;

    procedure tsymtable.foreach(proc2call : tcallback);

      procedure a(p : psym);

        { must be preorder, because it's used by reading in }
        { a PPU file                                        }
        begin
           proc2call(p);
           if assigned(p^.left) then a(p^.left);
           if assigned(p^.right) then a(p^.right);
        end;

      var
         i : integer;

      begin
         if hasharray<>nil then
           begin
              for i:=0 to hasharraysize-1 do
                if assigned(hasharray^[i]) then
                  a(hasharray^[i]);
           end
         else
           if assigned(wurzel) then
             a(wurzel);
      end;

    { write one symbol, is only used as call back procedure }
    procedure writesym(p : psym);far;

      begin
         p^.write;
      end;

    procedure tsymtable.writeasunit;

      var
         counter : word;
         s : string;
         p : psymtable;

      begin
         unitid:=0;

         { zuerst alle im Interface-Abschnitt aufgefhrten Units }
         { in die Datei schreiben und numerieren }
         p:=next;
         counter:=1;

         { im Implementationsteil aufgefuehrte Units ueberspringen }
         if symtabletype<>globalsymtable then
           begin
              while (p^.symtabletype<>globalsymtable) do
                p:=p^.next;
              p:=p^.next;
           end;
         while assigned(p) do
           begin
              if p^.symtabletype=unitsymtable then
                begin
                   p^.unitid:=counter;
                   inc(counter);
                   writebyte(ibloadunit);
                   writestring(p^.name^);
                   writelong(punitsymtable(p)^.checksum);
                end;
              p:=p^.next;
           end;

         { die Namen der benutzten Units schreiben }
         s:=usedunits^.get;
         while s<>'' do
           begin
              writebyte(ibinitunit);
              writestring(s);
              s:=usedunits^.get;
           end;
         s:=linkofiles.get;
         while s<>'' do
           begin
              writebyte(iblinkofile);
              writestring(s);
              s:=linkofiles.get;
           end;
         tsymtable.write;
      end;

    procedure tsymtable.writeasstruct;

      var
         counter : word;
         s : string;
         p : psymtable;

      begin
         tsymtable.write;
         {$IfDef GDB}
         writelong(call_offset);
         {$EndIf GDB}
      end;

    procedure tsymtable.write;

      var
         pd : pdef;
         counter : longint;

      begin
         { each definition get a number ... }
         counter:=0;
         pd:=wurzeldef;
         while assigned(pd) do
           begin
              pd^.number:=counter;
              inc(counter);
              pd:=pd^.next;
           end;
         { ...now write the definition }
         pd:=wurzeldef;
         while assigned(pd) do
           begin
              pd^.write;
              pd:=pd^.next;
           end;

         { the next part are the symbols }
         writebyte(ibend);

         { foreach is used to write all symbols }

         { FPKPascal needs a little bit other syntax }
{$ifdef tp}
         foreach(writesym);
{$else}
         foreach(@writesym);
{$endif}
         { end of symbols }
         writebyte(ibend);
      end;

{**************************************
              "forward"-pointer
 **************************************}

    type
       presolvelist = ^tresolvelist;

       tresolvelist = record
          p : ppointerdef;
          typ : ptypesym;
          next : presolvelist;
       end;

    var
       swurzel : presolvelist;

{$ifdef GDB}
    procedure clear_forwards;

      var
         p : presolvelist;

      begin
         p:=swurzel;
         while assigned(p) do
         begin
              swurzel:=p^.next;
            dispose(p);
            p := swurzel;
         end;
      end;

{$endif * GDB *}
    procedure save_forward(ppd : ppointerdef;typesym : ptypesym);

      var
         p : presolvelist;

      begin
         new(p);
         p^.next:=swurzel;
         p^.p:=ppd;
         p^.typ:=typesym;
         swurzel:=p;
      end;

    procedure resolve_forwards;

      var
         p : presolvelist;

      begin
         p:=swurzel;
         while p<>nil do
           begin
              swurzel:=swurzel^.next;
              p^.p^.definition:=p^.typ^.definition;
              dispose(p);
              p:=swurzel;
           end;
      end;

    constructor tsym.init(const n : string);

      begin
         left:=nil;
         right:=nil;
         setname(n);
         typ:=abstractsym;
         forwarddef:=false;
{$ifdef GDB}
         isstabwritten := false;
{$endif * GDB *}
      end;

    constructor tsym.load;

      begin
         left:=nil;
         right:=nil;
         setname(readstring);
         typ:=abstractsym;
         forwarddef:=false;
{$ifdef GDB}
         isstabwritten := false;
{$endif * GDB *}
      end;

    destructor tsym.done;

      begin
{$ifdef tp}
         if not(use_big) then
{$endif tp}
           strdispose(_name);
         if assigned(left) then dispose(left,done);
         if assigned(right) then dispose(right,done);
      end;

    procedure tsym.write;

      begin
         writestring(name);
      end;

    procedure tsym.deref;

      begin
      end;

    function tsym.name : string;

      var
         s : string;
         b : byte;

      begin
{$ifdef tp}
         if use_big then
           begin
              symbolstream.seek(longint(_name));
              symbolstream.read(b,1);
              symbolstream.read(s[1],b);
              s[0]:=chr(b);
              name:=s;
           end
         else
{$endif}
           begin
              name:=strpas(_name);
           end;
      end;

    procedure tsym.setname(const s : string);

      begin
         setstring(_name,s);
      end;

{$ifdef GDB}
    function tsym.stabstring : pchar;

      begin
         stabstring:=strpnew('"'+name+'",'+tostr(N_LSYM)+',0,0,0');
      end;

    procedure tsym.concatstabto(asmlist : paasmoutput);

      begin
         if not isstabwritten then
           begin
              asmlist^.concat(new(pai_stabs,init(stabstring)));
              isstabwritten:=true;
          end;
    end;
{$endif * GDB *}

{**************************************
               TLABELSYM
 **************************************}

    constructor tlabelsym.init(const n : string;l : longint);

      begin
         inherited init(n);
         typ:=labelsym;
         number:=l;
         defined:=false;
      end;

    destructor tlabelsym.done;

      begin
         if not(defined) then
           begin
              exterror:=strpnew(name);
              error(label_not_defined);
           end;
         inherited done;
      end;

    procedure tlabelsym.write;

      begin
         error(ill_label_pos);
      end;

{**************************************
               TUNITSYM
 **************************************}

    constructor tunitsym.init(const n : string;ref : psymtable);

      begin
         tsym.init(n);
         typ:=unitsym;
         unitsymtable:=ref;
      end;

    procedure tunitsym.write;

      begin
      end;

{$ifdef GDB}
    procedure tunitsym.concatstabto(asmlist : paasmoutput);
      begin
      {Nothing to write to stabs !}
      end;

{$endif * GDB *}
{$ifndef GDB}
{**************************************
               TTYPESYM
 **************************************}
{$else * GDB *}
{**************************************
               TERRORSYM
 **************************************}
{$endif * GDB *}

    constructor terrorsym.init;

      begin
         tsym.init('');
         typ:=errorsym;
      end;

{**************************************
               TVARSYM
 **************************************}

    constructor tvarsym.init(const n : string;p : pdef);

      begin
         tsym.init(n);
         typ:=varsym;
         definition:=p;
         varspez:=vs_value;
         adresse:=0;
         refs:=0;

         { can we load the value into a register ? }
         case p^.deftype of
            pointerdef,aufzaehldef,procvardef : regable:=true;
            grunddef : case pgrunddef(p)^.typ of
                          u8bit,s32bit,bool8bit,uchar,
                          s8bit,s16bit,u16bit,u32bit : regable:=true;
                          else regable:=false;
                       end;
            else regable:=false;
         end;
         reg:=R_NO;
      end;

    constructor tvarsym.load;

      begin
         tsym.load;
         typ:=varsym;
         varspez:=tvarspez(readbyte);
         if read_member then
           adresse:=readlong
         else adresse:=0;
         definition:=readdefref;

         { symbols which are load are never candidates for a register }
         regable:=false;
         reg:=R_NO;
      end;

    procedure tvarsym.deref;

      begin
         resolvedef(definition);
      end;

    procedure tvarsym.write;

      begin
         writebyte(ibvarsym);
         tsym.write;
         writebyte(byte(varspez));

         if read_member then
           writelong(adresse);

         writedefref(definition);
      end;

{$ifdef GDB}
    function tvarsym.stabstring : pchar;
    var st : char;
    begin
    {case varspez of}
    if owner^.symtabletype = globalsymtable then
      begin
      {Here we used S instead of G
      because with G GDB doesn't look at the address field
      but searches the same name or with a leading underscore
      but these names don't exist in pascal !}
      stabstring := strpnew('"'+name+':S'
               +definition^.numberstring+'",'+
               tostr(N_GSYM)+',0,0,'+'U_'+owner^.name^+'_'+name);
      {asmlist^.concat(gennasmrec(DIRECT,S_NO,'LG_'+name+':'));}
      end else
    if owner^.symtabletype = staticsymtable then
      begin
      stabstring := strpnew('"'+name+':S'
               +definition^.numberstring+'",'+
               tostr(N_LCSYM)+',0,0,'+'_'+name);
      {asmlist^.concat(gennasmrec(DIRECT,S_NO,'LS_'+name+':'));}
      end else
      if (owner^.symtabletype and parasymtable)<>0 then
        begin
        case varspez of
         vs_value : st := 'p';
         vs_var   : st := 'v';
         vs_const : st := 'i';
         end;
        stabstring := strpnew('"'+name+':'+st+
               +definition^.numberstring+'",'+
               tostr(N_PSYM)+',0,0,'+tostr(adresse+owner^.call_offset))
               {offset to ebp}
      end else
      if (owner^.symtabletype and localsymtable)<>0 then
        stabstring := strpnew('"'+name+':'
               +definition^.numberstring+'",'+
               tostr(N_LSYM)+',0,0,-'+tostr(adresse))
      else
      if owner^.symtabletype = unitsymtable then
        stabstring := strpnew('"'+name+':S'
               +definition^.numberstring+'",'+
               tostr(N_LCSYM)+',0,0,'+'U_'+owner^.name^+'_'+name)
      else
      stabstring := inherited stabstring;
    end;

{$endif * GDB *}
    function tvarsym.getsize : longint;

      begin
         { only if the definition is set, we could determine the   }
         { size, this is if an error occurs while reading the type }
         if assigned(definition) then
           begin
              case varspez of
                 vs_value : getsize:=definition^.size;
                 vs_var : getsize:=4;
                 vs_const : begin
                               if (definition^.deftype=stringdef) or
                                  (definition^.deftype=arraydef) or
                                  (definition^.deftype=recorddef) or
                                  (definition^.deftype=objectdef) or
                                  (definition^.deftype=setdef) then
                                  getsize:=4
                                else
                                  getsize:=definition^.size;
                            end;
              end;
           end;
      end;

{**************************************
               TTYPEDCONSTSYM
 **************************************}

    constructor ttypedconstsym.init(const n : string;p : pdef);

      begin
         tsym.init(n);
         typ:=typedconstsym;
         definition:=p;
         prefix:=stringdup(procprefix);
      end;

    constructor ttypedconstsym.load;

      begin
         tsym.load;
         typ:=typedconstsym;
         definition:=readdefref;
         prefix:=stringdup(readstring);
      end;

    destructor ttypedconstsym.done;

      begin
         stringdispose(prefix);
         tsym.done;
      end;

    procedure ttypedconstsym.deref;

      begin
         resolvedef(definition);
      end;

    procedure ttypedconstsym.write;

      begin
         writebyte(ibtypedconstsym);
         tsym.write;
         writedefref(definition);
         writestring(prefix^);
      end;

{$ifdef GDB}
    function ttypedconstsym.stabstring : pchar;
    var n : string;
    begin
    n :='TC_'+prefix^+'_'+name;
    stabstring := strpnew('"'+name+':S'
            +definition^.numberstring+'",'+tostr(n_STSYM)+',0,0,'+n);
    end;
{$endif * GDB *}
{**************************************
               TCONSTSYM
 **************************************}

    constructor tconstsym.init(const n : string;t : tconsttype;v : longint;def : pdef);

      begin
         tsym.init(n);
         typ:=constsym;
         definition:=def;
         consttype:=t;
         value:=v;
      end;

    constructor tconstsym.load;

      var
         pd : pdouble;

      begin
         tsym.load;
         typ:=constsym;
         consttype:=tconsttype(readbyte);
         case consttype of
            constint,
            constbool,
            constchar : value:=readlong;
            constord : begin
                          definition:=readdefref;
                          value:=readlong;
                       end;
            conststring : value:=longint(stringdup(readstring));
            constreal : begin
                           new(pd);
                           pd^:=readdouble;
                           value:=longint(pd);
                        end;
         else fatalerror(malformed_unit);
         end;
      end;

{$ifdef GDB}
    destructor tconstsym.done;
      begin
      if consttype = conststring then stringdispose(pstring(value));
      inherited done;
      end;

{$endif * GDB *}
    procedure tconstsym.deref;

      begin
         if consttype=constord then
           resolvedef(pdef(definition));
      end;

    procedure tconstsym.write;

      begin
         writebyte(ibconstsym);
         tsym.write;
         writebyte(byte(consttype));
         case consttype of
            constint,
            constbool,
            constchar : writelong(value);
            constord : begin
                          writedefref(definition);
                          writelong(value);
                       end;
            conststring : writestring(pstring(value)^);
            constreal : writedouble(pdouble(value)^);
            else internalerror(13);
         end;
      end;

{$ifdef GDB}
    function tconstsym.stabstring : pchar;
    var st : string;
    begin
         case consttype of
            constbool : st := 'b'+tostr(value);
          constord,
            constint : st := 'i'+tostr(value);
            constchar : st := 'c'+tostr(value);
            conststring : begin
                          st := ibm2ascii(pstring(value)^);
                          st := 's'''+st+'''';
                          end;

            constreal : begin
                        system.str(pdouble(value)^,st);
                        st := 'r'+st;
                        end;
          end;
    stabstring := strpnew('"'+name+':c='+st+'",'+tostr(N_function)+',0,0,0');
    end;

{$endif * GDB *}
{**************************************
               TAUFZAEHLSYM
 **************************************}

{$ifndef GDB}
    constructor taufzaehlsym.init(const n : string;def : paufzaehldef;v : longint);
{$else * GDB *}
    constructor taufzaehlsym.init(const n : string;def : paufzaehldef;v : longint;anext : paufzaehlsym);
{$endif * GDB *}

      begin
         tsym.init(n);
         typ:=aufzaehlsym;
         definition:=def;
         value:=v;
{$ifdef GDB}
         next := anext;
         def^.first := @self;
{$endif * GDB *}
      end;

    constructor taufzaehlsym.load;

      begin
         tsym.load;
         typ:=aufzaehlsym;
         definition:=paufzaehldef(readdefref);
         value:=readlong;
{$ifdef GDB}
         next := Nil;
{$endif * GDB *}
      end;

    procedure taufzaehlsym.deref;

      begin
         resolvedef(pdef(definition));
{$ifdef GDB}
         next := definition^.first;
         definition^.first := @self;
{$endif * GDB *}
      end;

    procedure taufzaehlsym.write;

      begin
         writebyte(ibaufzaehlsym);
         tsym.write;
         writedefref(definition);
         writelong(value);
{$ifdef GDB}
      end;

    procedure taufzaehlsym.concatstabto(asmlist : paasmoutput);
    begin
    {enum elements have no stab !}
{$endif * GDB *}
      end;

{$ifdef GDB}

{$endif * GDB *}
{**************************************
               TTYPESYM
 **************************************}

    constructor ttypesym.init(const n : string;d : pdef);

      begin
         tsym.init(n);
         typ:=typesym;
         definition:=d;

         { this allows to link definitions with the type with declares }
         { them                                                        }
         if assigned(definition) then
           if definition^.sym=nil then
             definition^.sym:=@self;
      end;

    constructor ttypesym.load;

      begin
         tsym.load;
         typ:=typesym;
         definition:=readdefref;
      end;

    destructor ttypesym.done;

      begin
         if assigned(definition) then
           if definition^.sym=@self then
             definition^.sym:=nil;
         inherited done;
      end;

    procedure ttypesym.deref;

      begin
         resolvedef(definition);
         if assigned(definition) then
           if definition^.sym=nil then
             definition^.sym:=@self;
      end;

    procedure ttypesym.write;

      begin
         writebyte(ibtypesym);
         tsym.write;
         writedefref(definition);
      end;

{$ifdef GDB}
    function ttypesym.stabstring : pchar;

      var
         stabchar : char;
         short : string;

      begin
         if definition^.deftype in [recorddef,setdef,aufzaehldef] then
           stabchar := 'T'
         else stabchar := 't';
         short:='"'+name+':'+stabchar+definition^.numberstring+
           '",'+tostr(N_LSYM)+',0,0,0';
         stabstring := strpnew(short);
      end;

    procedure ttypesym.concatstabto(asmlist : paasmoutput);

      begin
         if definition^.sym = @self then
           definition^.concatstabto(asmlist)
         else
           begin
              forcestabto(asmlist,definition);
              inherited concatstabto(asmlist);
           end;
      end;

{$endif * GDB *}

{**************************************
               TPROCSYM
 **************************************}

    procedure tprocsym.write;

      begin
         writebyte(ibprocsym);
         tsym.write;
         writedefref(pdef(definition));
      end;

    function tprocsym.is_global;

      begin
         internalerror(10003);
      end;

{$ifdef GDB}
    function tprocsym.stabstring : pchar;

      var
         rettype : char;
         obj,info : string;

      begin
         if is_global then
           rettype:='F'
         else
           rettype:='f';

         if owner^.symtabletype=objectsymtable then
           begin
              obj:=owner^.name^+'_';
              { rettype := '#'; }
           end
         else
           obj := '';
         if ((owner^.symtabletype and localsymtable)<>0) and
           assigned(owner^.name) then
           info:=','+name+','+owner^.name^
         else
           info := '';
         stabstring:=strpnew('"'+obj+name+':'+rettype+
           definition^.retdef^.numberstring+info+'",'+tostr(n_function)+
           ',0,0,'+definition^.mangledname);
      end;

    procedure tprocsym.concatstabto(asmlist : paasmoutput);

      begin
         if (definition^.options and pointernproc)<>0 then
           exit;

         asmlist^.concat(new(pai_stabs,init(stabstring)));

         isstabwritten := true;
         if assigned(definition^.parast) then
           definition^.parast^.concatstabto(asmlist);

         if assigned(definition^.localst) then
           definition^.localst^.concatstabto(asmlist);

         definition^.isstabwritten := true;
    end;

{$endif * GDB *}
{**************************************
               TSYSSYM
 **************************************}

    constructor tsyssym.init(const n : string;l : longint);

      begin
         inherited init(n);
         typ:=syssym;
         number:=l;
      end;

    procedure tsyssym.write;

      begin
      end;

{$ifdef GDB}
    procedure tsyssym.concatstabto(asmlist : paasmoutput);

      begin
      end;

{$endif * GDB *}
{**************************************
               TMACROSYM
 **************************************}

    constructor tmacrosym.init(const n : string);

      begin
         inherited init(n);
         defined:=true;
         buftext:=nil;
         buflen:=0;
      end;

    destructor tmacrosym.done;

      begin
         if assigned(buftext) then
           freemem(buftext,buflen);
         inherited done;
      end;

{$ifdef GDB}
    function globaldef(const s : string) : pdef;

      var st : string;
          symt : psymtable;
      begin
         srsym := nil;
         if pos('.',s) > 0 then
           begin
           st := copy(s,1,pos('.',s)-1);
           getsym(st,false);
           st := copy(s,pos('.',s)+1,255);
           if assigned(srsym) then
             begin
             if srsym^.typ = unitsym then
               begin
               symt := punitsym(srsym)^.unitsymtable;
               srsym := symt^.search(st);
               end else srsym := nil;
             end;
           end else st := s;
         if srsym = nil then getsym(st,true);
         if srsym^.typ<>typesym then
           begin
           error(type_id_expect);
           exit;
           end;
         globaldef := ptypesym(srsym)^.definition;
      end;

    function typeglobalnumber(const s : string) : string;

      var st : string;
          symt : psymtable;
      begin
         typeglobalnumber := '0';
         srsym := nil;
         if pos('.',s) > 0 then
           begin
           st := copy(s,1,pos('.',s)-1);
           getsym(st,false);
           st := copy(s,pos('.',s)+1,255);
           if assigned(srsym) then
             begin
             if srsym^.typ = unitsym then
               begin
               symt := punitsym(srsym)^.unitsymtable;
               srsym := symt^.search(st);
               end else srsym := nil;
             end;
           end else st := s;
         if srsym = nil then getsym(st,true);
         if srsym^.typ<>typesym then
           begin
           error(type_id_expect);
           exit;
           end;
         typeglobalnumber := ptypesym(srsym)^.definition^.numberstring;
      end;

{$endif * GDB *}
{**************************************
                  TDEF
 **************************************}

    constructor tdef.init;

      begin
         deftype:=abstractdef;
{$ifdef GDB}
         owner := nil;
       next := nil;
         globalnb := 0;
{$endif * GDB *}
         if registerdef then symtablestack^.registerdef(@self);
{$ifdef GDB}
         isstabwritten := false;
         isusedinstab := false;
         if assigned(lastglobaldef) then
           lastglobaldef^.nextglobal := @self
           else firstglobaldef := @self;
         lastglobaldef := @self;
         nextglobal := nil;
         sym := nil;
{$endif * GDB *}
      end;

{$ifdef GDB}
    constructor tdef.load;

      begin
         deftype:=abstractdef;
         isstabwritten := false;
         isusedinstab := false;
         globalnb := 0;
         if assigned(lastglobaldef) then
           lastglobaldef^.nextglobal := @self
           else firstglobaldef := @self;
         lastglobaldef := @self;
         nextglobal := nil;
         sym := nil;
         owner := nil;
         next := nil;
      end;

{$endif * GDB *}
    function tdef.size : longint;

      begin
         size:=savesize;
      end;

    procedure tdef.write;

      begin
      end;

{$ifdef GDB}
    function tdef.stabstring : pchar;

      begin
         stabstring:=strpnew('t'+numberstring+';');
      end;

    function tdef.numberstring : string;

      begin
         { formal def have no type !}
         if deftype = formaldef then
           begin
              numberstring:=voiddef^.numberstring;
              exit;
           end;
         if not(isusedinstab) then
           begin
              { set even if debuglist is not defined }
              isusedinstab := true;
              if assigned(debuglist) and not isstabwritten then
               concatstabto(debuglist);
           end;
         if globalnb = 0 then
           begin
              globalnb:=globaltypecount;
              inc(globaltypecount);
           end;
         numberstring := tostr(globalnb);
      end;

    function tdef.allstabstring : pchar;

      var
         stabchar : char;
         ss,st : pchar;
         name : string;

      begin
         ss:=stabstring;
         getmem(st,strlen(ss)+512);
         if deftype in [recorddef,setdef,aufzaehldef] then
           stabchar:= 'T'
         else stabchar := 't';
{$ifndef GDBknowsStrings}
         { GDB 4.12 for go32 doesn't know strings }
         if deftype = stringdef then stabchar := 'T';
{$endif}
      {$IfNDef GDBknowsfiles}
      { GDB 4.12 for go32 doesn't know files }
      if deftype = stringdef then stabchar := 'T';
      {$EndIf}
      if assigned(sym) then
        name := sym^.name
        else name := ' ';
      strpcopy(st,'"'+name+':'+stabchar+numberstring+'=');
      strpcopy(strecopy(strend(st),ss),'",'+tostr(N_LSYM)+',0,0,0');
      allstabstring := strnew(st);
      freemem(st,strlen(ss)+512);
      strdispose(ss);
      end;


    procedure tdef.concatstabto(asmlist : paasmoutput);

      begin
         if isusedinstab and not isstabwritten then
           begin
              asmlist^.concat(new(pai_stabs,init(allstabstring)));
              isstabwritten := true;
           end;
      end;

{$endif * GDB *}
    procedure tdef.deref;

      begin
      end;

    destructor tdef.done;
{$ifndef GDB}

{$else * GDB *}
      var pd : pdef;
      begin
      pd := firstglobaldef;
      if pd = @self then firstglobaldef := pd^.nextglobal
        else while assigned(pd) do
{$endif * GDB *}
      begin
{$ifdef GDB}
            if pd^.nextglobal = @Self then
               begin
               pd^.nextglobal := pd^.nextglobal^.nextglobal;
               if pd^.nextglobal = nil then lastglobaldef := pd;
               end;
            pd := pd^.nextglobal;
            end;
{$endif * GDB *}
      end;

{**************************************
              TSTRINGDEF
 **************************************}

    constructor tstringdef.init(l : byte);

      begin
         tdef.init;
         deftype:=stringdef;
         len:=l;
         savesize:=len+1;
      end;

    constructor tstringdef.load;

      begin
{$ifdef GDB}
         tdef.load;
{$endif * GDB *}
         deftype:=stringdef;
         len:=readbyte;
         savesize:=len+1;
      end;

    procedure tstringdef.write;

      begin
         writebyte(ibstringdef);
         writebyte(len);
      end;

{$ifdef GDB}
    function tstringdef.stabstring : pchar;
      var bytest,charst : string;
      begin
      charst := typeglobalnumber('SYSTEM.CHAR');
      { this is what I found in stabs.texinfo but
      gdb 4.12 for go32 doesn't understand that !! }
      {$IfDef GDBknowsstrings}
      stabstring := strpnew('n'+charst+';'+tostr(len));
      {$else}
      bytest := typeglobalnumber('SYSTEM.BYTE');
      stabstring := strpnew('s'+tostr(len+1)+'length:'+bytest
                    +',0,8;st:ar'+bytest
                    +';1;'+tostr(len)+';'+charst+',8,'+tostr(len*8)+';;');
      {$EndIf}
      end;

    procedure tstringdef.concatstabto(asmlist : paasmoutput);
      var systemsymt : psymtable;
      begin
      if isusedinstab and not isstabwritten then
        begin
        if cs_compilesystem in aktswitches then
          systemsymt := symtablestack
          else
          begin
          getsym('SYSTEM',true);
          if srsym^.typ = unitsym then systemsymt := punitsym(srsym)^.unitsymtable
            else systemsymt := Nil;
          end;
        {$IFNDef GDBknowsstrings}
      if assigned(systemsymt) then
        srsym := systemsymt^.search('BYTE')
        else getsym('BYTE',false);
        if assigned(srsym) and (srsym^.typ = typesym) then
           srsym^.concatstabto(asmlist)
           else forcestabto(asmlist,u8bitdef);
        {$EndIf}
      if assigned(systemsymt) then
        srsym := systemsymt^.search('CHAR')
        else getsym('CHAR',false);
        getsym('CHAR',false);
        if assigned(srsym) and (srsym^.typ = typesym) then
           srsym^.concatstabto(asmlist)
           else forcestabto(asmlist,cchardef);
        inherited concatstabto(asmlist);
        end;
      end;
{$endif * GDB *}
{**************************************
             TAUFZAEHLDEF
 **************************************}

    constructor taufzaehldef.init;

      begin
         tdef.init;
         deftype:=aufzaehldef;
         max:=0;
         savesize:=4;
{$ifdef GDB}
         first := Nil;
{$endif * GDB *}
      end;

    constructor taufzaehldef.load;

      begin
{$ifdef GDB}
         tdef.load;
{$endif * GDB *}
         deftype:=aufzaehldef;
         max:=readlong;
         savesize:=4;
{$ifdef GDB}
         first := Nil;
      end;

    destructor taufzaehldef.done;

      begin
         inherited done;
{$endif * GDB *}
      end;

    procedure taufzaehldef.write;

      begin
         writebyte(ibaufzaehldef);
         writelong(max);
{$ifdef GDB}
      end;

    function taufzaehldef.stabstring : pchar;
      var st,st2 : pchar;
          p : paufzaehlsym;
          s : string;
          memsize : word;
      begin
      memsize := 2048;
      getmem(st,memsize);
      strpcopy(st,'e');
      p := first;
      while assigned(p) do
        begin
        s :=p^.name+':'+tostr(p^.value)+',';
        if (strlen(st)+length(s)<memsize) then
          strpcopy(strend(st),s)
          else
          begin
          getmem(st2,memsize+2048);
          strcopy(st2,st);
          freemem(st,memsize);
          st := st2;
          memsize := memsize+2048;
          strpcopy(strend(st),s);
          end;
        p := p^.next;
        end;
      strpcopy(strend(st),';');
      stabstring := strnew(st);
      freemem(st,memsize);
{$endif * GDB *}
      end;

{**************************************
               TGRUNDDEF
 **************************************}

    constructor tgrunddef.init(t : tbasetyp;v,b : longint);

      begin
         tdef.init;
         deftype:=grunddef;
         von:=v;
         bis:=b;
         typ:=t;
         setsize;
      end;

    constructor tgrunddef.load;

      begin
{$ifdef GDB}
         tdef.load;
{$endif * GDB *}
         deftype:=grunddef;
         typ:=tbasetyp(readbyte);
         von:=readlong;
         bis:=readlong;
         setsize;

         { set voiddef to a definition from system }
         if (typ=uvoid) and not(cs_compilesystem in aktswitches) then
{$ifdef GDB}
           begin
           dispose(voiddef,done);
{$endif * GDB *}
           voiddef:=@self;
{$ifdef GDB}
           voidpointerdef^.definition := voiddef;
         end;
{$endif * GDB *}
      end;

    procedure tgrunddef.setsize;

      begin
         if typ=uauto then
           begin
              if (von>=0) and (bis<=255) then
                begin
                   savesize:=1;
                   typ:=u8bit;
                end
              else if (von>=-128) and (bis<=127) then
                begin
                   savesize:=1;
                   typ:=s8bit;
                end
              else if (von>=0) and (bis<=65536) then
                begin
                   savesize:=2;
                   typ:=u16bit;
                end
              else if (von>=-32768) and (bis<=32767) then
                begin
                   savesize:=2;
                   typ:=s16bit;
                end
              else
                begin
                   savesize:=4;
                   typ:=s32bit;
                end;
           end
         else
           case typ of
              uchar,u8bit,bool8bit,s8bit : savesize:=1;
              u16bit,s16bit : savesize:=2;
              s32bit,u32bit  : savesize:=4;
              s64real : savesize:=8;
              else savesize:=0;
           end;

         { there are no entrys for range checking }
         rangenr:=0;
      end;

    procedure tgrunddef.genrangecheck;

      var
         rangelabel : tlabel;

      begin
         if rangenr=0 then
           begin
              { generate two constant for bounds }
              getlabel(rangenr);
              rangenr:=rangelabel;
              datasegment^.concat(new(pai_symbol,init('R_'+tostr(rangenr))));
              datasegment^.concat(new(pai_const,init_32bit(von)));
              datasegment^.concat(new(pai_const,init_32bit(bis)));
           end;
      end;

    procedure tgrunddef.write;

      begin
         writebyte(ibgrunddef);
         writebyte(byte(typ));
         writelong(von);
         writelong(bis);
      end;

{$ifdef GDB}
    function tgrunddef.stabstring : pchar;

      begin
      case typ of
         s64real : stabstring:=strpnew('r'+
              s32bitdef^.numberstring+';'+tostr(savesize)+';0;');

         uvoid : stabstring:=strpnew(numberstring+';');
         {GDB 4.12 for go32 doesn't like boolean as range for 0 to 1 !!!}
         bool8bit : stabstring:=strpnew('r'+numberstring+';0;255;');
         u32bit : stabstring:=strpnew('r'+
              s32bitdef^.numberstring+';0;-1;');
         else stabstring:=strpnew('r'+s32bitdef^.numberstring+';'
            +tostr(von)+';'+tostr(bis)+';');
         end;
      end;

{$endif * GDB *}
{**************************************
               TFILEDEF
 **************************************}

    constructor tfiledef.init(ft : tfiletyp;tas : pdef);

      begin
         inherited init;
         deftype:=filedef;
         filetyp:=ft;
         typed_as:=tas;
         setsize;
      end;

    constructor tfiledef.load;

      begin
{$ifdef GDB}
         inherited load;
{$endif * GDB *}
         deftype:=filedef;
         filetyp:=tfiletyp(readbyte);
         if filetyp=ft_typed then
           typed_as:=readdefref
         else
           typed_as:=nil;
         setsize;
      end;

    procedure tfiledef.deref;

      begin
         if filetyp=ft_typed then
           resolvedef(typed_as);
      end;

    procedure tfiledef.write;

      begin
         writebyte(ibfiledef);
         writebyte(byte(filetyp));
         if filetyp=ft_typed then
           writedefref(typed_as);
      end;

{$ifdef GDB}
    function tfiledef.stabstring : pchar;

      begin
      {$IfDef GDBknowsfiles}
      case filetyp of
        ft_typed : stabstring := strpnew('d'+typed_as^.numberstring{+';'});
        ft_untyped : stabstring := strpnew('d'+voiddef^.numberstring{+';'});
        ft_text : stabstring := strpnew('d'+cchardef^.numberstring{+';'});
        end;
      {$Else }
      {based on
       filerec = record
          handle : word;
          mode : word;
          recsize : word;
          _private : array[1..26] of byte;
          userdata : array[1..16] of byte;
          name : string[79];}

      stabstring := strpnew('s128HANDLE:'+typeglobalnumber('SYSTEM.WORD')+',0,16;'+
                      'MODE:'+typeglobalnumber('SYSTEM.WORD')+',16,16;'+
                      'RECSIZE:'+typeglobalnumber('SYSTEM.WORD')+',32,16;'+
                      '_PRIVATE:ar'+typeglobalnumber('SYSTEM.WORD')+';1;26;'+typeglobalnumber('SYSTEM.BYTE')+',36,208;'+
                      'USERDATA:ar'+typeglobalnumber('SYSTEM.WORD')+';1;16;'+typeglobalnumber('SYSTEM.BYTE')+',256,128;'+
                      'NAME:s80'+
                        'length:'+typeglobalnumber('SYSTEM.BYTE')+',0,8;'+
                        'st:ar'+typeglobalnumber('SYSTEM.WORD')+';1;79;'+typeglobalnumber('SYSTEM.CHAR')+',8,632;;'+
                      ',384,640;;');
      {$EndIf}
      end;

    procedure tfiledef.concatstabto(asmlist : paasmoutput);

      begin
      if isusedinstab and not isstabwritten then
        begin
        if assigned(typed_as) then forcestabto(asmlist,typed_as);
        inherited concatstabto(asmlist);
        end;
      end;

{$endif * GDB *}
    procedure tfiledef.setsize;

      begin
         if target_info.target=target_LINUX then
           begin
              case filetyp of
                 ft_text : savesize:=432;
                 ft_untyped : savesize:=304;
                 {!!!!!!}
                 else internalerror(17);
              end;
           end
         else
           begin
              case filetyp of
                 ft_text : savesize:=256;
                 ft_untyped : savesize:=128;
                 {!!!!!!}
                 else internalerror(17);
              end;
           end;
      end;

{**************************************
               TPOINTERDEF
 **************************************}

    constructor tpointerdef.init(def : pdef);

      begin
         inherited init;
         deftype:=pointerdef;
         definition:=def;
         savesize:=4;
      end;

    constructor tpointerdef.load;

      begin
{$ifdef GDB}
         tdef.load;
{$endif * GDB *}
         deftype:=pointerdef;
         { die Adressen werden spter berechnet }
         definition:=readdefref;
         savesize:=4;
      end;

    procedure tpointerdef.deref;

      begin
         resolvedef(definition);
      end;

    procedure tpointerdef.write;

      begin
         writebyte(ibpointerdef);
         writedefref(definition);
      end;

{$ifdef GDB}
    function tpointerdef.stabstring : pchar;

      begin
      stabstring := strpnew('*'+definition^.numberstring);
      end;

    procedure tpointerdef.concatstabto(asmlist : paasmoutput);

      var
         st : string;

      begin
         if isusedinstab and not isstabwritten then
           begin
              if assigned(definition) then
                if definition^.deftype in [recorddef,objectdef] then
                  begin
                     if not definition^.isstabwritten then
                       begin
                          if assigned(definition^.sym) then
                            begin
                               if assigned(sym) then
                                 st:=sym^.name
                               else
                                 st:=' ';

                               st:='"'+st+':t'+numberstring+'=*'+definition^.numberstring
                                 +'=xs'+definition^.sym^.name+':",'+tostr(N_LSYM)+',0,0,0';

                               asmlist^.concat(new(pai_stabs,init(strpnew(st))));
                            end;
                       end
                     else
                       inherited concatstabto(asmlist);
                     isstabwritten := true;
                  end
                else
                  begin
                     forcestabto(asmlist,definition);
                     inherited concatstabto(asmlist);
                  end;
           end;
      end;

{$endif * GDB *}
{**************************************
               TSETDEF
 **************************************}

    constructor tsetdef.init(s : pdef;high : longint);

      begin
         inherited init;
         deftype:=setdef;
         setof:=s;
         { nur normale Sets mit 32 Bytes Gre werden momentan untersttzt }
         if high<256 then
           begin
              settyp:=normset;
              savesize:=32;
           end
         else error(illsettype);
      end;

    constructor tsetdef.load;

      begin
{$ifdef GDB}
         tdef.load;
{$endif * GDB *}
         deftype:=setdef;
         setof:=readdefref;
         settyp:=tsettyp(readbyte);
         savesize:=32;
      end;

    procedure tsetdef.write;

      begin
         writebyte(ibsetdef);
         writedefref(setof);
         writebyte(byte(settyp));
      end;

{$ifdef GDB}
    function tsetdef.stabstring : pchar;

      begin
      stabstring := strpnew('S'+setof^.numberstring+';');
      end;

    procedure tsetdef.concatstabto(asmlist : paasmoutput);

      begin
      if isusedinstab and not isstabwritten then
        begin
        if assigned(setof) then forcestabto(asmlist,setof);
        inherited concatstabto(asmlist);
        end;
      end;

{$endif * GDB *}
    procedure tsetdef.deref;

      begin
         resolvedef(setof);
      end;

{**************************************
               TFORMALDEF
 **************************************}

    constructor tformaldef.init;

      begin
         inherited init;
         deftype:=formaldef;
         savesize:=4;
      end;

    constructor tformaldef.load;

      begin
{$ifdef GDB}
         tdef.load;
{$endif * GDB *}
         deftype:=formaldef;
         savesize:=4;
      end;

    procedure tformaldef.write;

      begin
         writebyte(ibformaldef);
      end;

{$ifdef GDB}
    function tformaldef.stabstring : pchar;

      begin
      stabstring := strpnew('formal'+numberstring+';');
      end;


    procedure tformaldef.concatstabto(asmlist : paasmoutput);

      begin
      { formaldef can't be stab'ed !}
      end;
{$endif * GDB *}
{**************************************
               TARRAYDEF
 **************************************}

    constructor tarraydef.init(l,h : longint;rd : pdef);

      begin
         tdef.init;
         deftype:=arraydef;
         lowrange:=l;
         highrange:=h;
         rangedef:=rd;
         rangenr:=0;
      end;

    constructor tarraydef.load;

      begin
{$ifdef GDB}
         tdef.load;
{$endif * GDB *}
         deftype:=arraydef;
         rangenr:=0;

         { die Adressen werden spter berechnet }
         definition:=readdefref;
         rangedef:=readdefref;
         lowrange:=readlong;
         highrange:=readlong;
      end;

    procedure tarraydef.genrangecheck;

      var
         rangelabel : tlabel;

      begin
         if rangenr=0 then
           begin
              { generates the data for range checking }

              { the following construction is stupid, but for future use }
              getlabel(rangelabel);
              rangenr:=rangelabel;
              datasegment^.concat(new(pai_symbol,init('R_'+tostr(rangenr))));
              datasegment^.concat(new(pai_const,init_32bit(lowrange)));
              datasegment^.concat(new(pai_const,init_32bit(highrange)));
           end;
      end;

    procedure tarraydef.deref;

      begin
         resolvedef(definition);
         resolvedef(rangedef);
      end;

    procedure tarraydef.write;

      begin
         writebyte(ibarraydef);
         writedefref(definition);
         writedefref(rangedef);
         writelong(lowrange);
         writelong(highrange);
      end;

{$ifdef GDB}
    function tarraydef.stabstring : pchar;
      begin
      stabstring := strpnew('ar'+{rangedef^.numberstring}typeglobalnumber('SYSTEM.LONGINT')+';'
                    +tostr(lowrange)+';'+tostr(highrange)+';'+definition^.numberstring);
      end;

    procedure tarraydef.concatstabto(asmlist : paasmoutput);

      begin
      if isusedinstab and not isstabwritten then
        begin
        {if assigned(rangedef) then forcestabto(asmlist,rangedef);}
        if assigned(definition) then forcestabto(asmlist,definition);
        inherited concatstabto(asmlist);
        end;
      end;

{$endif * GDB *}
    function tarraydef.elesize : longint;

      begin
         elesize:=definition^.size;
      end;

    function tarraydef.size : longint;

      begin
         size:=(highrange-lowrange+1)*elesize;
      end;

{**************************************
               TRECDEF
 **************************************}

    constructor trecdef.init(p : psymtable);

      begin
         tdef.init;
         deftype:=recorddef;
         symtable:=p;
         savesize:=symtable^.datasize;
      end;

    constructor trecdef.load;

      var
         oldread_member : boolean;

      begin
{$ifdef GDB}
         tdef.load;
{$endif * GDB *}
         deftype:=recorddef;
         savesize:=readlong;
         oldread_member:=read_member;
         read_member:=true;
         symtable:=new(psymtable,loadasstruct(recordsymtable));
         read_member:=oldread_member;
      end;

    destructor trecdef.done;

      begin
{$ifndef GDB}
         dispose(symtable);
{$else * GDB *}
         if assigned(symtable) then dispose(symtable,done);
         inherited done;
{$endif * GDB *}
      end;

    procedure derefsym(p : psym);far;

      begin
         p^.deref;
      end;

    procedure trecdef.deref;

      var
         hp : pdef;
         oldrecsyms : psymtable;

      begin
         oldrecsyms:=aktrecordsymtable;
         aktrecordsymtable:=symtable;
         { nun die Definitionen Dereferenzieren }
         hp:=symtable^.wurzeldef;
         while assigned(hp) do
           begin
              hp^.deref;

              {Besitzer setzten }
              hp^.owner:=symtable;

              hp:=hp^.next;
           end;
{$ifdef tp}
         symtable^.foreach(derefsym);
{$else}
         symtable^.foreach(@derefsym);
{$endif}
         aktrecordsymtable:=oldrecsyms;
      end;

    procedure trecdef.write;

      var
         oldread_member : boolean;

      begin
         oldread_member:=read_member;
         read_member:=true;
         writebyte(ibrecorddef);
         writelong(savesize);
         self.symtable^.writeasstruct;
         read_member:=oldread_member;
      end;

{$ifdef GDB}

      Const StabRecString : pchar = Nil;
          RecOffset : Longint = 0;
          Objecthasvirtual : boolean = False;

    procedure addname(p :psym);far;
    var newst,virtualind : string;
        savestab, newrec : pchar;
        pd,ipd : pprocdef;
        lindex : longint;
    begin
    If p^.typ = varsym then
       begin
       newrec := strpnew(p^.name+':'+pvarsym(p)^.definition^.numberstring
                     +','+tostr(pvarsym(p)^.adresse*8)+','
                     +tostr(pvarsym(p)^.definition^.size*8)+';');
       strcat(StabRecstring,newrec);
       strdispose(newrec);
       { this should be used for case! }
       recoffset:=recoffset+pvarsym(p)^.definition^.size;
       end;
    if p^.typ = procsym then
      if (pprocsym(p)^.definition^.options and povirtualmethod) <> 0 then
        objecthasvirtual := true;
    end;

    function trecdef.stabstring : pchar;

      var
         oldrec : pchar;

      begin
         oldrec := stabrecstring;
         getmem(stabrecstring,2048);
         strpcopy(stabRecString,'s'+tostr(savesize));
         recoffset:=0;
{$ifdef tp}
         symtable^.foreach(addname);
{$else}
         symtable^.foreach(@addname);
{$endif}
          { !!!! fpk doesn't want to convert a char to a pchar}
          { is this a bug ? }
          strpcopy(strend(StabRecString),';');
          stabstring:=strnew(StabRecString);
          freemem(stabrecstring,2048);
          stabrecstring:=oldrec;
      end;

    procedure trecdef.concatstabto(asmlist : paasmoutput);

      begin
      if isusedinstab and not isstabwritten then
        begin
        inherited concatstabto(asmlist);
        end;
      end;

{$endif * GDB *}
{**************************************
               TABSTRACTPROCDEF
 **************************************}

    constructor tabstractprocdef.init;

      begin
         inherited init;
         para1:=nil;
         options:=0;
         retdef:=voiddef;
         savesize:=4;
      end;

    destructor tabstractprocdef.done;

      var
         hp : pdefcoll;

      begin
         hp:=para1;
         while assigned(hp) do
           begin
              para1:=hp^.next;
              dispose(hp);
              hp:=para1;
           end;
         inherited done;
      end;

    procedure tabstractprocdef.concatdef(p : pdef;vsp : tvarspez);

      var
         hp : pdefcoll;

      begin
         new(hp);
         hp^.paratyp:=vsp;
         hp^.data:=p;
         hp^.next:=para1;
         para1:=hp;
      end;

    procedure tabstractprocdef.deref;

      var
         hp : pdefcoll;

      begin
         inherited deref;
         resolvedef(retdef);
         hp:=para1;
         while assigned(hp) do
           begin
              resolvedef(hp^.data);
              hp:=hp^.next;
           end;
      end;

    constructor tabstractprocdef.load;

      var
         last,hp : pdefcoll;
         count,i : word;

      begin
{$ifdef GDB}
         inherited load;
{$endif}
         retdef:=readdefref;
         options:=readword;
         count:=readword;
         para1:=nil;
         savesize:=4;
         for i:=1 to count do
           begin
              new(hp);
              hp^.paratyp:=tvarspez(readbyte);
              hp^.data:=readdefref;
              hp^.next:=nil;
              if para1=nil then
                para1:=hp
              else
                last^.next:=hp;
              last:=hp;
           end;
      end;

    procedure tabstractprocdef.write;

      var
         count : word;
         hp : pdefcoll;

      begin
         writedefref(retdef);
         writeword(options);
         hp:=para1;
         count:=0;
         while assigned(hp) do
           begin
              inc(count);
              hp:=hp^.next;
           end;
         writeword(count);
         hp:=para1;
         while assigned(hp) do
           begin
              writebyte(byte(hp^.paratyp));
              writedefref(hp^.data);
              hp:=hp^.next;
           end;
      end;

{$ifdef GDB}
    function tabstractprocdef.stabstring : pchar;
      begin
      stabstring := strpnew('abstractproc'+numberstring+';');
      end;

    procedure tabstractprocdef.concatstabto(asmlist : paasmoutput);

      begin
      if isusedinstab and not isstabwritten then
        begin
        {if assigned(retdef) then forcestabto(asmlist,retdef);}
        inherited concatstabto(asmlist);
        end;
      end;

{$endif * GDB *}
{**************************************
               TPROCDEF
 **************************************}

    constructor tprocdef.init;

      begin
         inherited init;
         deftype:=procdef;
         _mangledname:=nil;
         nextoverloaded:=nil;
         extnumber:=-1;
         parast:=new(psymtable,init(parasymtable));
         localst:=new(psymtable,init(localsymtable));
         { grundstzlich sind alle Register benutzt    }
         { erst wenn eine Implementation vorliegt wird }
         { dies vom Parser gegebenenfalls gendert     }
         usedregisters:=$ff;
         forwarddef:=true;
{$ifdef GDB}
         _class := nil;
{$endif * GDB *}
      end;

    constructor tprocdef.load;

      var
         s : string;

      begin
         deftype:=procdef;
         inherited load;

         usedregisters:=readbyte;
         s:=readstring;
         setstring(_mangledname,s);

         { this symbol is external: }
         externals^.concat(new(pai_external,init(s)));

         extnumber:=readlong;
         nextoverloaded:=pprocdef(readdefref);
{$ifdef GDB}
         _class := pobjectdef(readdefref);
{$endif * GDB *}

         if gendeffile and ((options and poexports)<>0) then
           writeln(defdatei,#9+mangledname);

         parast:=nil;
         localst:=nil;
         forwarddef:=false;
      end;

    destructor tprocdef.done;

      var
         hp : pdefcoll;

      begin
         if assigned(parast) then
           dispose(parast,done);
         if assigned(localst) then
           dispose(localst,done);
         if
{$ifdef tp}
         not(use_big) and
{$endif}
         assigned(_mangledname) then
           strdispose(_mangledname);
         inherited done;
      end;

    procedure tprocdef.write;

      begin
         writebyte(ibprocdef);
         inherited write;

         writebyte(usedregisters);
         writestring(mangledname);
         writelong(extnumber);
         writedefref(nextoverloaded);
{$ifdef GDB}
         writedefref(_class);
      end;

    procedure addparaname(p : psym);far;
      var vs : char;

      begin
      if pvarsym(p)^.varspez = vs_value then vs := '1'
        else vs := '0';
      strpcopy(strend(StabRecString),p^.name+':'+pvarsym(p)^.definition^.numberstring+','+vs+';');
      end;

    function tprocdef.stabstring : pchar;
      var param : pdefcoll;
          i : word;
          vartyp : char;
          oldrec : pchar;
      begin
      oldrec := stabrecstring;
      getmem(StabRecString,1024);
      param := para1;
      i := 0;
      while assigned(param) do
        begin
        inc(i);
        param := param^.next;
        end;
      strpcopy(StabRecString,'f'+retdef^.numberstring);
      if i>0 then
        begin
        strpcopy(strend(StabRecString),','+tostr(i)+';');
        if assigned(parast) then
{$ifdef TP}
          parast^.foreach(addparaname)
{$else}
          parast^.foreach(@addparaname)
{$endif}
          else
          begin
          param := para1;
          i := 0;
          while assigned(param) do
            begin
            inc(i);
            if param^.paratyp = vs_value then
              vartyp := '1' else vartyp := '0';
            { Here we have lost the parameter names !!}
            strpcopy(strend(stabrecstring),'PARAM'+tostr(i)
               +':'+param^.data^.numberstring+','+vartyp+';');
            param := param^.next;
            end;
          end;
        {strpcopy(strend(StabRecString),';');}
        end;
      stabstring := strnew(stabrecstring);
      freemem(stabrecstring,1024);
      stabrecstring := oldrec;
      end;

    procedure tprocdef.concatstabto(asmlist : paasmoutput);

      begin
      if (options and pointernproc) <> 0 then exit;
      if isusedinstab and not isstabwritten then
        begin
        if assigned(nextoverloaded) then
          nextoverloaded^.concatstabto(asmlist)
          else inherited concatstabto(asmlist);
        isstabwritten := true;
        end;
{$endif * GDB *}
      end;

    procedure tprocdef.deref;

      begin
         inherited deref;
         resolvedef(pdef(nextoverloaded));
{$ifdef GDB}
         resolvedef(pdef(_class));
{$endif * GDB *}
      end;

    function tprocdef.mangledname : string;

      var
         oldpos : longint;
         s : string;
         b : byte;

      begin
{ $ifdef tp
         if use_big then
           begin
              symbolstream.seek(longint(_mangledname));
              symbolstream.read(b,1);
              symbolstream.read(s[1],b);
              s[0]:=chr(b);
              mangledname:=s;
           end
         else
$endif}
           begin
              mangledname:=strpas(_mangledname);
           end;
      end;

    procedure tprocdef.setmangledname(const s : string);

      begin
         if
{$ifdef tp}
         not(use_big) and
{$endif}
         (assigned(_mangledname)) then
           strdispose(_mangledname);
         setstring(_mangledname,s);
      end;

{**************************************
               TPROCVARDEF
 **************************************}

    constructor tprocvardef.init;

      begin
         inherited init;
         deftype:=procvardef;
      end;

    constructor tprocvardef.load;

      begin
{$ifndef GDB}
         deftype:=procvardef;
{$endif * not GDB *}
         inherited load;
{$ifdef GDB}
         deftype:=procvardef;
{$endif * GDB *}
      end;

    procedure tprocvardef.write;

      begin
         writebyte(ibprocvardef);
         inherited write;
      end;

{$ifdef GDB}
    function tprocvardef.stabstring : pchar;

      var nss : pchar;
          i : word;
          vartyp : char;
          pst : pchar;
          param : pdefcoll;
      begin
      i := 0;
      param := para1;
      while assigned(param) do
        begin
        inc(i);
        param := param^.next;
        end;
      getmem(nss,1024);
      strpcopy(nss,'f'+retdef^.numberstring+','+tostr(i)+';');
      param := para1;
      i := 0;
      while assigned(param) do
        begin
        inc(i);
        if param^.paratyp = vs_value then vartyp := '1' else vartyp := '0';
        {Here we have lost the parameter names !!}
        pst := strpnew('PARAM'+tostr(i)+':'+param^.data^.numberstring+','+vartyp+';');
        strcat(nss,pst);
        strdispose(pst);
        param := param^.next;
        end;
      {strpcopy(strend(nss),';');}
      stabstring := strnew(nss);
      freemem(nss,1024);
      end;

{$endif * GDB *}
{**************************************
               TOBJECTDEF
 **************************************}

{$ifdef GDB}
 Const vtabletype : word = 0;
       vtableassigned : boolean = False;

{$endif * GDB *}
   constructor tobjectdef.init(const n : string;c : pobjectdef);

     begin
        tdef.init;
        deftype:=objectdef;
        childof:=c;
        options:=0;
        {privatesyms:=new(psymtable,init(objectsymtable));
        protectedsyms:=new(psymtable,init(objectsymtable)); }
        publicsyms:=new(psymtable,init(objectsymtable));
{$ifndef GDB}

{$else * GDB *}
        publicsyms^.name := stringdup(n);
{$endif * GDB *}
        { Daten der Vorfahren bei den Adressen beachten }
        if assigned(childof) then
          publicsyms^.datasize:=
            publicsyms^.datasize-4+childof^.publicsyms^.datasize;
        name:=stringdup(n);
{$ifdef GDB}
        deftype:=objectdef;
        savesize := publicsyms^.datasize;
{$endif * GDB *}
     end;

    constructor tobjectdef.load;

      var
         oldread_member : boolean;

      begin
{$ifdef GDB}
         tdef.load;
{$endif * GDB *}
         deftype:=objectdef;
         savesize:=readlong;
         name:=stringdup(readstring);

         { defines the VMT external }
         externals^.concat(new(pai_external,init('VMT_'+name^)));

         childof:=pobjectdef(readdefref);
         options:=readword;
         oldread_member:=read_member;
         read_member:=true;
         publicsyms:=new(psymtable,loadasstruct(objectsymtable));
         publicsyms^.datasize:=savesize;
{$ifdef GDB}
         publicsyms^.name := stringdup(name^);
{$endif * GDB *}
         read_member:=oldread_member;

         { handles the predefined class tobject  }
         { the last TOBJECT which is loaded gets }
         { it !                                  }
         if (name^='TOBJECT') and not(cs_compilesystem in aktswitches) and
           ( (options and oois_class)<>0 ) and (childof=pointer($ffffffff)) then
           class_tobject:=@self;
{$ifdef GDB}
         deftype:=objectdef;
{$endif * GDB *}
      end;

   destructor tobjectdef.done;

     begin
{!!!!
        if assigned(privatesyms) then
          dispose(privatesyms,done);
        if assigned(protectedsyms) then
          dispose(protectedsyms,done); }
        if assigned(publicsyms) then
          dispose(publicsyms,done);
        stringdispose(name);
        tdef.done;
     end;

   function tobjectdef.isrelated(d : pobjectdef) : boolean;

     var
        hp : pobjectdef;

     begin
        isrelated:=false;
        hp:=@self;
        while assigned(hp) do
          begin
             if hp=d then
               begin
                  isrelated:=true;
                  exit;
               end;
             hp:=hp^.childof;
          end;
     end;

   function tobjectdef.size : longint;

     begin
        if (options and oois_class)<>0 then
          size:=4
        else
          size:=publicsyms^.datasize;
     end;

    procedure tobjectdef.deref;

      var
         hp : pdef;
         oldrecsyms : psymtable;

      begin
         resolvedef(pdef(childof));
         oldrecsyms:=aktrecordsymtable;
         aktrecordsymtable:=publicsyms;
         { nun die Definitionen dereferenzieren }
         hp:=publicsyms^.wurzeldef;
         while assigned(hp) do
           begin
              hp^.deref;

              {Besitzer setzen }
              hp^.owner:=publicsyms;

              hp:=hp^.next;
           end;
{$ifdef tp}
         publicsyms^.foreach(derefsym);
{$else}
         publicsyms^.foreach(@derefsym);
{$endif}
         aktrecordsymtable:=oldrecsyms;
      end;

    procedure tobjectdef.write;

      var
         oldread_member : boolean;

      begin
         oldread_member:=read_member;
         read_member:=true;
         writebyte(ibobjectdef);
         writelong(size);
         writestring(name^);
         writedefref(childof);
         writeword(options);
         publicsyms^.writeasstruct;
         read_member:=oldread_member;
      end;

{$ifdef GDB}
    procedure addprocname(p :psym);far;

      var
         newst,virtualind : string;
         savestab, newrec : pchar;
         pd,ipd : pprocdef;
         lindex : longint;

    begin
    If p^.typ = procsym then
       begin
                pd := pprocsym(p)^.definition;
                { this will be used for full implementation of object stabs
                not yet done }
                ipd := pd;
                while assigned(ipd^.nextoverloaded) do
                  ipd := ipd^.nextoverloaded;
                if (pd^.options and povirtualmethod) <> 0 then
                   begin
                   objecthasvirtual := true;
                   lindex := pd^.extnumber;
                   lindex := lindex or $80000000;
                   virtualind := '*'+tostr(lindex)+';'+ipd^._class^.numberstring
                   end else virtualind := '.';
                { arguments are not listed here }
                {we don't need another definition}
                ipd^.isstabwritten := true;
                newrec := strpnew(p^.name+'::'+ipd^.numberstring
                     +'=##'+pd^.retdef^.numberstring+';:;2A'
                     +virtualind+';');
               strcat(StabRecstring,newrec);
               strdispose(newrec);
               {This should be used for case !!}
               RecOffset := RecOffset + pd^.size;
       end;
    end;

    function tobjectdef.stabstring : pchar;
      var anc : pobjectdef;
          oldohv : boolean;
          oldrec : pchar;
      begin
      oldohv := objecthasvirtual;
      objecthasvirtual := false;
      oldrec := stabrecstring;
      GetMem(stabrecstring,2048);
      strpcopy(stabRecString,'s'+tostr(size));
      if assigned(childof) then
        {only one ancestor not virtual, public, at base offset 0 }
        {       !1           ,    0       2         0    ,       }
        strpcopy(strend(stabrecstring),'!1,020,'+childof^.numberstring+';');
      {virtual table to implement yet}
      RecOffset := 0;
{$ifdef tp}
         publicsyms^.foreach(addname);
         if objecthasvirtual then
            strpcopy(strend(stabrecstring),'$vf'+numberstring+':'+tostr(vtabletype)+',0;');
         publicsyms^.foreach(addprocname);
{$else}
         publicsyms^.foreach(@addname);
         if objecthasvirtual then
            strpcopy(strend(stabrecstring),'$vf'+numberstring+':'+tostr(vtabletype)+',0;');
         publicsyms^.foreach(@addprocname);
{$endif}
      if objecthasvirtual then
        begin
        anc := @self;
        while assigned(anc^.childof) do anc := anc^.childof;
        strpcopy(strend(stabRecString),'~%'+anc^.numberstring+';');
        end;
      stabstring := strnew(StabRecString);
      Freemem(stabrecstring,2048);
      stabrecstring := oldrec;
      objecthasvirtual := oldohv;
      end;

    procedure tobjectdef.concatstabto(asmlist : paasmoutput);

      begin

         if isusedinstab and not isstabwritten then
           begin
              if not vtableassigned then
                begin
                   vtabletype:=globalTypeCount;
                   inc(globalTypeCount,3);
                   asmlist^.concat(new(pai_stabs,init(strpnew('"$vtbl_ptr_type:Tt'+tostr(vtabletype+1)+'=s16'+
                     'delta:'+typeglobalnumber('SYSTEM.LONGINT')+',0,32;'+
                     'moinsdelta:'+typeglobalnumber('SYSTEM.LONGINT')+',32,32;'+
                     'parent_vmt:'+typeglobalnumber('SYSTEM.LONGINT')+',64,32;'+
                     'pfn:'+typeglobalnumber('SYSTEM.POINTER')+',96,32;;",128,0,0,0'))));
                   asmlist^.concat(new(pai_stabs,init(strpnew('" :t'+tostr(vtabletype)+'=*'+
                     tostr(vtabletype+2)+'=ar'+typeglobalnumber('SYSTEM.INTEGER')+';0;1;'+tostr(vtabletype+1)+
                       '",128,0,0,0'))));
                   vtableassigned:=true;
                end;
              if assigned(childof) then
                forcestabto(asmlist,childof);
              inherited concatstabto(asmlist);
              asmlist^.concat(new(pai_stabs,init(strpnew('"'+name^+':T'+numberstring+'",'+
                tostr(N_LSYM)+',0,0,0'))));
              isstabwritten:=true;
           end;
      end;
{$endif * GDB *}

{**************************************
               TERRORDEF
 **************************************}

   constructor terrordef.init;

     begin
        tdef.init;
        deftype:=errordef;
     end;

{$ifdef GDB}
    function terrordef.stabstring : pchar;

      begin
      stabstring := strpnew('error'+numberstring);
      end;


{$endif * GDB *}
   procedure init_symtable;

     begin
        macros:=new(psymtable,init(macrosymtable));
        usedunits:=nil;
        read_member:=false;
        generrorsym:=new(perrorsym,init);
        swurzel:=nil;
        { readunit_lastloaded:=nil; }
        loaded_units:=nil;
{$ifdef GDB}
        firstglobaldef := nil;
        lastglobaldef := nil;
{$endif * GDB *}
     end;

   procedure done_symtable;

     var    hp : ploaded_units;

     begin
        dispose(macros,done);
        dispose(generrorsym,done);
        while assigned(symtablestack) do dellexlevel;
        { !!!!!
        clear_forwards;
        }
        while assigned(loaded_units) do
          begin
          hp := loaded_units;
          loaded_units := hp^.next;
          {name are disposed by symtables}
          dispose(hp);
     end;
         dispose(generrordef,done);
         dispose(s32bitdef,done);
         {$IfNDef TestString}
         dispose(cstringdef,done);
         {$EndIf TestString }
         dispose(cchardef,done);
         dispose(cs64realdef,done);
         { !!!!!! u32bit ?}

         { sonstige verwendete Definitionen: }
         {dispose(voiddef,done); belongs to system !}
         dispose(u8bitdef,done);
         dispose(u16bitdef,done);
         dispose(booldef,done);
         dispose(voidpointerdef,done);
     end;

{ management for units }

{**************************************
               TUNIT
**************************************}

    constructor tunit.init(const p,n : string;anext : punit);

      begin
         symtab:=nil;
         next:=anext;
         path:=stringdup(p);
         name:=stringdup(n);
         sourcename:=nil;
         sourceext:=nil;
      end;

    function searchcompile : punit;

      var
         hp : punit;

      begin
         searchcompile:=nil;
         hp:=units;
         while assigned(hp) do
           begin
              if hp^.symtab=nil then
                begin
                   searchcompile:=hp;
                   exit;
                end;
              hp:=hp^.next;
           end;
      end;

    function searchunit(const s : string) : punit;

      var
         hp : punit;

      begin
         searchunit:=nil;
         hp:=units;
         while assigned(hp) do
           begin
              if upper(hp^.name^)=upper(s) then
                begin
                   searchunit:=hp;
                   exit;
                end;
              hp:=hp^.next;
           end;
      end;

    function readunit(const s : string) : punitsymtable;

      var
         unitfile : file;
         dummy,source_avail : boolean;
         unitfilesize : longint;
         b : byte;
         hp : punit;
         oldbuffer : prbuffer;
         oldbufferl,oldbufferp : longint;
         checksum,unit_time,source_time : longint;
         unit_path,source_path : string;

      begin
         hp:=searchunit(s);
         if assigned(hp) then
           begin
              { try already to load the unit ? }
              if (hp^.symtab=nil) then
                { we can't recover from such an error }
                fatalerror(rec_unit_def)
              else
                begin
                   readunit:=hp^.symtab;
                   exit;
                end;
           end;
         oldbuffer:=rbuffer;
         oldbufferp:=bufferp;
         oldbufferl:=bufferl;
{$ifdef linux}
         from:=globals.lowercase(from);
{$endif}

         { unit isn't loaded }
         { try to load the unit }

         { search first the source }
         source_path:=search(s+target_info.sourceext,inputdir,source_avail);
         if not(source_avail) then
           source_path:=search(s+target_info.sourceext,unitsearchpath,source_avail);

         { search the directory of the input file }
         unit_path:=search(s+target_info.unitext,inputdir,dummy);
         if not(dummy) then
           unit_path:=search(s+target_info.unitext,unitsearchpath,dummy);

         if not(dummy) then
           begin
              { we must in any case compile the unit }
              { if we don't find the PPU file        }
              if source_avail then
                begin
                   { source is available          }
                   { so we could compile the unit }
                   units:=new(punit,init(source_path,s,units));
                   units^.sourcename:=stringdup(s);
                   units^.sourceext:=stringdup(target_info.sourceext);
                   readunit:=nil;
                   exit;
                end
              else
                begin
                   exterror:=strpnew(s+target_info.unitext);
                   { I think, it's seamless to try to recover from }
                   { this error                                    }
                   fatalerror(unit_not_found);
                end;
           end
         else
           begin
              { if we have a PPU and a PP file: }
              if source_avail then
                begin
                   { get file times }
                   assign(unitfile,unit_path+s+target_info.unitext);
                   getftime(unitfile,unit_time);
                   assign(unitfile,source_path+s+target_info.sourceext);
                   getftime(unitfile,source_time);

                   { compile the source:         }
                   { if there is a build command }
                   if do_build or
                   { the source and the unit are at different locations }
                   (unit_path<>source_path) or
                   { the source is newer than the unit file }
                     (source_time>unit_time) then
                     begin
                        units:=new(punit,init(source_path,s,units));
                        units^.sourcename:=stringdup(s);
                        units^.sourceext:=stringdup(target_info.sourceext);
                        readunit:=nil;
                        exit;
                     end;
                end;
           end;
         assign(unitfile,unit_path+s+target_info.unitext);
         reset(unitfile,1);
         units:=new(punit,init(unit_path,s,units));

         unitfilesize:=filesize(unitfile);

         if unitfilesize>maxunitsize then
           fatalerror(error_reading_unit);

         { maybe write a message }
         if veryverbose then
           writeln('Loading '+unit_path+target_info.unitext)
         else if not(quiet) then
           writeln('Loading '+s+target_info.unitext);

         getmem(rbuffer,unitfilesize);
         blockread(unitfile,rbuffer^,unitfilesize,bufferl);
         close(unitfile);

         { check the format }
         for b:=0 to 5 do
           if rbuffer^[b]<>byte(unitheader[b]) then
             fatalerror(malformed_unit);

         { check the target OS }
         if (rbuffer^[8])<>byte(target_info.target) then
           fatalerror(not_same_target);

         bufferp:=sizeof(unitheader);

         units^.symtab:=new(punitsymtable,load(s));
         readunit:=units^.symtab;

         { read the checksum of the file }
         checksum:=plongint(@rbuffer^[10])^;
         units^.symtab^.maschstart:=plongint(@rbuffer^[16])^;
         freemem(rbuffer,unitfilesize);

         rbuffer:=oldbuffer;
         bufferp:=oldbufferp;
         bufferl:=oldbufferl;
      end;

begin
   units:=nil;
end.
