{****************************************************************************

                   Copyright (c) 1993,96 by Florian Klaempfl

 ****************************************************************************}

unit tgeni386;

  interface

    uses
       cobjects,globals,errors,tree,hcodegen
{$ifdef i386}
       ,i386
       ,aasm
{$endif}
       ;

    const
       usablereg32 : byte = 4;

    function getregister32 : tregister;
    procedure ungetregister32(r : tregister);

    procedure ungetregister(r : tregister);

    procedure cleartempgen;

    { generates temporary variables }
    procedure resettempgen;
    procedure setfirsttemp(l : longint);
    function gettempsize : longint;
{$ifndef Tempad}
    function gettempofsize(l : longint) : longint;
{$else * Tempad *}
    function gettempofsize(size : longint) : longint;
{$endif * Tempad *}
    procedure gettempofsizereference(l : longint;var ref : treference);
{$ifndef Tempad}
    procedure ungettemp(pos : longint;s : longint);
{$else * Tempad *}
    function istemp(var ref : treference) : boolean;
    procedure ungetiftemp(var ref : treference);
    {procedure ungettemp(pos : longint;size : longint);
    function gettempstr(l : longint) : string; }
{$endif * Tempad *}

    procedure del_reference(const ref : treference);

    type
       tregisterset = set of tregister;

    var
       unused,usableregs : tregisterset;
{$ifndef Tempad}
       firsttemp,c_usableregs : longint;
{$else * Tempad *}
       c_usableregs : longint;
{$endif * Tempad *}

       { Pseudomenge }
       usedinproc : byte;

       { count, how much a register must be pushed if it is used as register }
       { variable                                                            }
       reg_pushes : array[R_EAX..R_EDI] of longint;

  implementation

    procedure ungetregister(r : tregister);

      begin
         if r in [R_EAX,R_ECX,R_EDX,R_EBX,R_ESP,R_EBP,R_ESI,R_EDI] then
           ungetregister32(r)
         else if r in [R_AX,R_CX,R_DX,R_BX,R_SP,R_BP,R_SI,R_DI] then
           ungetregister32(reg16toreg32(r))
         else if r in [R_AL,R_BL,R_CL,R_DL] then
           ungetregister32(reg8toreg32(r))
         else internalerror(18);
      end;

    procedure ungetregister32(r : tregister);

      begin
         if cs_maxoptimieren in aktswitches then
           begin
              { takes much time }
              if not(r in usableregs) then
                exit;
              unused:=unused+[r];
              inc(usablereg32);
           end
         else
           begin
              if not(r in [R_EAX,R_EBX,R_ECX,R_EDX]) then
                exit;
              unused:=unused+[r];
              inc(usablereg32);
           end;
      end;

    procedure del_reference(const ref : treference);

      begin
         if ref.isintvalue then
           exit;
         ungetregister32(ref.base);
         ungetregister32(ref.index);
      end;

    function getregister32 : tregister;

      begin
         dec(usablereg32);
         if R_EAX in unused then
           begin
              unused:=unused-[R_EAX];
              usedinproc:=usedinproc or ($80 shr byte(R_EAX));
              getregister32:=R_EAX;
           end
         else if R_EDX in unused then
           begin
              unused:=unused-[R_EDX];
              usedinproc:=usedinproc or ($80 shr byte(R_EDX));
              getregister32:=R_EDX;
           end
         else if R_EBX in unused then
           begin
              unused:=unused-[R_EBX];
              usedinproc:=usedinproc or ($80 shr byte(R_EBX));
              getregister32:=R_EBX;
           end
         else if R_ECX in unused then
           begin
              unused:=unused-[R_ECX];
              usedinproc:=usedinproc or ($80 shr byte(R_ECX));
              getregister32:=R_ECX;
           end
         else internalerror(10);
      end;

    procedure cleartempgen;

      begin
         unused:=usableregs;
         usablereg32:=c_usableregs;
      end;

{$ifdef Tempad}
        type
       pfreerecord = ^tfreerecord;

       tfreerecord = record
          next : pfreerecord;
          pos : longint;
          size : longint;
       end;


    var freelist : pfreerecord;
        templist : pfreerecord;
        lastoccupied : longint;
        firsttemp, maxtemp : longint;

{$endif * Tempad *}
    procedure resettempgen;

      begin
{$ifndef Tempad}
         firsttemp:=0;
{$else * Tempad *}
         setfirsttemp(0);
{$endif * Tempad *}
      end;

    procedure setfirsttemp(l : longint);

{$ifdef Tempad}
    var hp : pfreerecord;

{$endif * Tempad *}
      begin
         firsttemp:=l;
{$ifdef Tempad}
         maxtemp := l;

         while assigned(freelist) do
           begin
           hp := freelist;
           freelist := hp^.next;
           dispose(hp);
{$endif * Tempad *}
      end;
{$ifdef Tempad}
         while assigned(templist) do
           begin
           {$IfDef ExtDebug}
           exterror := strpnew('temporary assignment of size '
                       +tostr(templist^.size)+
                       +' at pos '+tostr(templist^.pos)+
                       ' not freed at the end of the procedure');
           warning(user_defined);
           {$EndIf}
           hp := templist;
           templist := hp^.next;
           dispose(hp);
           end;
        lastoccupied := l;
      end;

{$endif * Tempad *}

{$ifndef Tempad}
    function gettempofsize(l : longint) : longint;
{$else * Tempad *}
    function gettempofsize(size : longint) : longint;
      var
         last,hp : pfreerecord;
{$endif * Tempad *}

      begin
{$ifndef Tempad}
         dec(firsttemp,l);
         if (firsttemp mod 4)<>0 then
           dec(firsttemp,4-(firsttemp mod 4));
         gettempofsize:=firsttemp;
{$else * Tempad *}
         if (size mod 4)<>0 then
           size:=size+(4-(size mod 4));
           if assigned(freelist) then
             begin
                last:=nil;
                hp:=freelist;
                while assigned(hp) do
                  begin
                     { erster passender Block wird genommen }
                     if hp^.size>=size then
                       begin
                          gettempofsize:=hp^.pos;
                          if hp^.pos-size < maxtemp then
                            maxtemp := hp^.size-size;
                          { wird der ganze Block bentigt ? }
                          if hp^.size>size then
                            begin
                               hp^.size:=hp^.size-size;
                               hp^.pos:=hp^.pos-size;
                            end
                          else
                            begin
                               if assigned(last) then
                                 last^.next:=hp^.next
                               else
                                 freelist:=nil;
                               dispose(hp);
                            end;
                          exit;
                       end;
                     last:=hp;
                     hp:=hp^.next;
                  end;
             end;
          { nothing free is big enough : expand temp }
          gettempofsize:=lastoccupied-size;
          lastoccupied:=lastoccupied-size;
          if lastoccupied < maxtemp then maxtemp := lastoccupied;

{$endif * Tempad *}
      end;

    function gettempsize : longint;

      begin
{$ifndef Tempad}
         gettempsize:=-firsttemp;
{$else * Tempad *}
         gettempsize:=-maxtemp;
{$endif * Tempad *}
      end;

{$ifndef Tempad}
    procedure ungettemp(pos : longint;s : longint);
{$else * Tempad *}

    function gettempstr(l : longint) : string;
{$endif * Tempad *}

      begin
{$ifndef Tempad}
         { !!!!!! Nur Dummy, wird spter vebessert }
{$else * Tempad *}
         gettempstr:=tostr(l)+'('+reg2str(procinfo.framepointer)+')';
{$endif * Tempad *}
      end;

    procedure gettempofsizereference(l : longint;var ref : treference);
{$ifndef Tempad}

{$else * Tempad *}
         var tl : pfreerecord;
{$endif * Tempad *}
      begin
         clear_reference(ref);
         ref.offset:=gettempofsize(l);
         ref.base:=procinfo.framepointer;
{$ifdef Tempad}
         ref.symbol:=nil;
         new(tl);
         tl^.pos := ref.offset;
         tl^.size := l;
         tl^.next := templist;
         templist := tl;
      end;

    function istemp(var ref : treference) : boolean;
      begin
      istemp := ((ref.base = procinfo.framepointer) and (ref.offset < firsttemp));
      end;

    procedure ungettemp(pos : longint;size : longint);

      var
         hp,newhp : pfreerecord;

      begin
         if (size mod 4)<>0 then
           size:=size+(4-(size mod 4));
         if size = 0 then exit;
         if pos<=lastoccupied then
           if pos = lastoccupied
            then lastoccupied:=pos+size
            else
            begin
            {$IfDef ExtDebug}
            exterror := strpnew('temp managment problem : ungettemp pos < lastoccupied !');
            warning(user_defined);
            {$EndIf}
            end
         else
           begin
              new(newhp);
              { size can be allways set }
              newhp^.size:=size;
              newhp^.pos := pos;
              { if there is no free list }
              if not assigned(freelist) then
                begin
                   { then generate one }
                   freelist:=newhp;
                   newhp^.next:=nil;
                   exit;
                end;
              { an welcher Position der freelist einfgen? }
              hp:=freelist;
              while assigned(hp) do
                begin
                   { conneting two blocks ? }
                   if hp^.pos+hp^.size=pos then
                      begin
                         inc(hp^.size,size);
                         dispose(newhp);
                         break;
                      end
                   { if the end is reached, then concat }
                   else if hp^.next=nil then
                     begin
                        hp^.next:=newhp;
                        newhp^.next:=nil;
                        break;
                     end
                   { falls der nchste Zeiger grer ist, dann }
                   { Einhngen                                 }
                   else if hp^.next^.pos<=pos+size then
                     begin
                        { vielleicht zwei Blcke zusammenfassen ? }
                        if pos+size=hp^.next^.pos then
                          begin
                             newhp^.next:=hp^.next^.next;
                             inc(newhp^.size,hp^.next^.size);
                             dispose(hp^.next);
                          end
                        else
                          begin
                             newhp^.next:=hp^.next;
                             hp^.next:=newhp;
                          end;
                        break;
                     end;
                   hp:=hp^.next;
                end;
           end;
      end;


    procedure ungetiftemp(var ref : treference);
      var tl,prev : pfreerecord;
      begin
      If istemp(ref) then
        begin
        prev := Nil;
        tl := templist;
        while assigned(tl) do
          begin
          if ref.offset = tl^.pos then
            begin
            ungettemp(ref.offset,tl^.size);
            if assigned(prev) then
              prev^.next := tl^.next
              else templist := tl^.next;
            dispose(tl);
            exit;
            end else
            begin
            prev := tl;
            tl := tl^.next;
            end;
          end;
        {$IfDef ExtDebug}
        exterror := strpnew('temp managment problem : temp ont found for release !');
        warning(user_defined);
        {$EndIf}
        end;
{$endif * Tempad *}
      end;

begin
   usableregs:=[R_EAX,R_EBX,R_ECX,R_EDX];
   c_usableregs:=4;
{$ifdef Tempad}
   freelist := nil;
   templist := Nil;
{$endif * Tempad *}
end.
